#include <stdio.h>
#include <stdlib.h>

#include <windows.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spDirectory.h>
#include <sp/spLibrary.h>
#include <sp/spPluginP.h>

#include <sp/spDIBitmapP.h>

#include "spSusie.h"

#define SP_SUSIE_INFO_BUFFER_SIZE 1024

#define SP_NUM_SUSIE_TYPE 1
#define SP_SUSIE_TYPE_DEFAULT 0

static spDIBitmapPluginRec *spGetSusieDIBitmapPluginRec(void);

typedef struct _spSusieSPI *spSusieSPI;

struct _spSusieSPI {
    char *name;
    void *handle;

    spSusieGetPluginInfoProc getPluginInfo;
    spSusieIsSupportedProc isSupported;
    spSusieGetPictureInfoProc getPictureInfo;
    spSusieGetPictureProc getPicture;
    
    int pluginTypeCount;
    
    int typeOffset;
    int typeCount;
    
    spSusieSPI next;
};

typedef struct _spSusiePluginInstance
{
    char type[SP_MAX_LINE];
    int typeIndex;

    spSusieSPI spi;
    char *filename;
} *spSusiePluginInstance;

static spSusieSPI sp_susie_first_spi = NULL;
static spSusieSPI sp_susie_last_spi = NULL;

static int sp_susie_num_file_type_list = 0;
static char **sp_susie_file_type_list = NULL;
static char **sp_susie_file_desc_list = NULL;
static char **sp_susie_file_filter_list = NULL;
    
int spRegisterSusieFileExtensions(const char *extensions, const char *fileType)
{
    int count;
    int extLength;
    int fileTypeLength;
    char *ptr;
    char *nextPtr;

    spDebug(80, "spRegisterSusieFileExtensions", "in: sp_susie_num_file_type_list = %d\n",
	    sp_susie_num_file_type_list);
    
    fileTypeLength = (int)strlen(fileType);

    count = 0;
    ptr = (char *)extensions;
    do {
	spDebug(100, "spRegisterSusieFileExtensions", "count = %d, ptr = %s\n", count, ptr);
	nextPtr = spStrChr(ptr, ';');
	if (nextPtr == NULL) {
	    extLength = (int)strlen(ptr);
	} else {
	    extLength = (int)(nextPtr - ptr);
	    ++nextPtr;
	}
	spDebug(100, "spRegisterSusieFileExtensions", "count = %d, extLength = %d, sp_susie_num_file_type_list = %d\n",
		count, extLength, sp_susie_num_file_type_list);
	
	sp_susie_file_type_list = xrealloc(sp_susie_file_type_list, sp_susie_num_file_type_list + 2, char *);
	sp_susie_file_desc_list = xrealloc(sp_susie_file_desc_list, sp_susie_num_file_type_list + 2, char *);
	sp_susie_file_filter_list = xrealloc(sp_susie_file_filter_list, sp_susie_num_file_type_list + 2, char *);
	
	sp_susie_file_filter_list[sp_susie_num_file_type_list] = xalloc(extLength + 1, char);
	memcpy(sp_susie_file_filter_list[sp_susie_num_file_type_list], ptr, extLength);
	sp_susie_file_filter_list[sp_susie_num_file_type_list][extLength] = NUL;
	spDebug(100, "spRegisterSusieFileExtensions", "filter = %s\n", sp_susie_file_filter_list[sp_susie_num_file_type_list]);
	
	sp_susie_file_desc_list[sp_susie_num_file_type_list] = xalloc(fileTypeLength + extLength + 2, char);
	spStrCopy(sp_susie_file_desc_list[sp_susie_num_file_type_list], fileTypeLength + 1, fileType);
	spStrCat(sp_susie_file_desc_list[sp_susie_num_file_type_list], fileTypeLength + 2, " ");
	spStrCat(sp_susie_file_desc_list[sp_susie_num_file_type_list], fileTypeLength + 2 + extLength,
		 sp_susie_file_filter_list[sp_susie_num_file_type_list]);

	if ((ptr = spStrChr(sp_susie_file_filter_list[sp_susie_num_file_type_list], '.')) != NULL) {
	    sp_susie_file_type_list[sp_susie_num_file_type_list] = strclone(ptr + 1);
	} else {
	    sp_susie_file_type_list[sp_susie_num_file_type_list] = strclone(sp_susie_file_filter_list[sp_susie_num_file_type_list]);
	}

	++sp_susie_num_file_type_list;
	++count;
	ptr = nextPtr;
    } while (ptr != NULL);

    sp_susie_file_type_list[sp_susie_num_file_type_list] = NULL;
    sp_susie_file_desc_list[sp_susie_num_file_type_list] = NULL;
    sp_susie_file_filter_list[sp_susie_num_file_type_list] = NULL;

    {
	spDIBitmapPluginRec *prec;
	prec = spGetSusieDIBitmapPluginRec();
	
	prec->file_type_list = (const char **)sp_susie_file_type_list;
	prec->file_desc_list = (const char **)sp_susie_file_desc_list;
	prec->file_filter_list = (const char **)sp_susie_file_filter_list;
    }

    spDebug(80, "spRegisterSusieFileExtensions", "done: sp_susie_num_file_type_list = %d\n",
	    sp_susie_num_file_type_list, count);
    
    return count;
}

void spFreeSusieFileTypeList(void)
{
    int i;

    spDebug(100, "spFreeSusieFileTypeList", "sp_susie_num_file_type_list = %d\n",
	    sp_susie_num_file_type_list);
    
    for (i = 0; i < sp_susie_num_file_type_list; i++) {
	xfree(sp_susie_file_type_list[i]);
	xfree(sp_susie_file_desc_list[i]);
	xfree(sp_susie_file_filter_list[i]);
    }
    
    xfree(sp_susie_file_type_list);
    xfree(sp_susie_file_desc_list);
    xfree(sp_susie_file_filter_list);

    sp_susie_file_type_list = NULL;
    sp_susie_file_desc_list = NULL;
    sp_susie_file_filter_list = NULL;
    sp_susie_num_file_type_list = 0;
    
    {
	spDIBitmapPluginRec *prec;
	prec = spGetSusieDIBitmapPluginRec();
	
	prec->file_type_list = NULL;
	prec->file_desc_list = NULL;
	prec->file_filter_list = NULL;
    }
    
    spDebug(100, "spFreeSusieFileTypeList", "done\n");
    
    return;
}

void spUpdateSusieFileTypeList(spSusieSPI spi)
{
    int n;
    int size;
    char extensions[SP_SUSIE_INFO_BUFFER_SIZE];
    char fileType[SP_SUSIE_INFO_BUFFER_SIZE];

    spi->typeOffset = sp_susie_num_file_type_list;
    spi->typeCount = 0;
    spi->pluginTypeCount = 0;
    
    for (n = 2;; n += 2) {
	spDebug(100, "spUpdateSusieFileTypeList", "n = %d\n", n);
	if ((size = spi->getPluginInfo(n, extensions, sizeof(extensions))) <= 0) {
	    break;
	}
	extensions[size] = NUL;
	spDebug(100, "spUpdateSusieFileTypeList", "extensions = %s\n", extensions);

	if ((size = spi->getPluginInfo(n + 1, fileType, sizeof(fileType))) <= 0) {
	    break;
	}
	fileType[size] = NUL;
	spDebug(100, "spUpdateSusieFileTypeList", "fileType = %s\n", fileType);

	spi->typeCount += spRegisterSusieFileExtensions(extensions, fileType);
	
	spi->pluginTypeCount++;
    }

    return;
}

spSusieSPI spCreateSusieSPI(const char *filename)
{
    void *handle;
    char buf[256];
    spSusieGetPluginInfoProc getPluginInfo;
    spSusieIsSupportedProc isSupported;
    spSusieGetPictureInfoProc getPictureInfo;
    spSusieGetPictureProc getPicture;
    spSusieSPI spi = NULL;

    spDebug(80, "spCreateSusieSPI", "filename = %s\n", filename);
    
    if ((handle = spOpenLibrary(filename)) != NULL) {
	if ((getPluginInfo = spGetSymbolAddress(handle, "GetPluginInfo")) != NULL
	    && getPluginInfo(0, buf, sizeof(buf)) >= 4 && strneq(buf, "00IN", 4)) {
	    isSupported = spGetSymbolAddress(handle, "IsSupported");
	    getPictureInfo = spGetSymbolAddress(handle, "GetPictureInfo");
	    getPicture = spGetSymbolAddress(handle, "GetPicture");

	    if (isSupported != NULL && getPictureInfo != NULL && getPicture != NULL) {
		spi = xalloc(1, struct _spSusieSPI);
		spi->name = strclone(filename);
		spi->handle = handle;
		spi->getPluginInfo = getPluginInfo;
		spi->isSupported = isSupported;
		spi->getPictureInfo = getPictureInfo;
		spi->getPicture = getPicture;
		spi->pluginTypeCount = 0;
		spi->typeOffset = 0;
		spi->typeCount = 0;
		spi->next = NULL;
		if (sp_susie_first_spi == NULL) {
		    sp_susie_first_spi = spi;
		} else {
		    sp_susie_last_spi->next = spi;
		}
		sp_susie_last_spi = spi;

		spUpdateSusieFileTypeList(spi);
	    } else {
		spDebug(10, "spCreateSusieSPI", "cannot obtain some symbols\n");
	    }
	} else {
	    spDebug(10, "spCreateSusieSPI", "get symbol address failed\n");
	}

	if (spi == NULL) {
	    spCloseLibrary(handle);
	}
    } else {
	spDebug(10, "spCreateSusieSPI", "spOpenLibrary failed\n");
    }
    
    return spi;
}

void spDestroySusieSPI(spSusieSPI spi)
{
    if (spi != NULL) {
	if (spi->handle != NULL) {
	    spCloseLibrary(spi->handle);
	}
	if (spi->name != NULL) {
	    xfree(spi->name);
	}

	xfree(spi);
    }

    return;
}

void spDestroyAllSusieSPI(void)
{
    if (sp_susie_first_spi != NULL) {
	spSusieSPI spi, nspi;

	spi = sp_susie_first_spi;
	while (spi != NULL) {
	    nspi = spi->next;
	    spDestroySusieSPI(spi);
	    spi = nspi;
	}
	sp_susie_first_spi = NULL;
    }

    return;
}

static spBool spInitPluginSusie(const char *lang)
{
    static char current_dir[SP_MAX_PATHNAME] = "";
    static char plugin_dir[SP_MAX_PATHNAME] = "";

    if (sp_susie_first_spi == NULL) {
	GetCurrentDirectory(sizeof(current_dir), current_dir);
	spDebug(80, "spInitPluginSusie", "current_dir = %s\n", current_dir);
    
	if (GetModuleFileName(NULL, plugin_dir, sizeof(plugin_dir))) {
	    char drive[SP_MAX_PATHNAME];
	    char dir[SP_MAX_PATHNAME];
	    DIR *dirp = NULL;
	    struct dirent *dp;
	    spSusieSPI spi = NULL;
	    
#if !defined(__CYGWIN32__)
	    _splitpath(plugin_dir, drive, dir, NULL, NULL);
	    strcpy(plugin_dir, drive);
	    strcat(plugin_dir, dir);
#else
	    spGetDirName(plugin_dir);
	    strcat(plugin_dir, SP_DIR_SEPARATOR_STRING);
#endif
	    strcat(plugin_dir, "plugins");
	    SetCurrentDirectory(plugin_dir);

	    spDebug(80, "spInitPluginSusie", "plugin_dir = %s\n", plugin_dir);
	    
	    if ((dirp = spOpenDir(plugin_dir)) != NULL) {
		while (1) {
		    if ((dp = spReadDir(dirp)) == NULL) {
			spDebug(80, "spInitPluginSusie", "spReadDir: no more file\n");
			break;
		    }
		    spDebug(80, "spInitPluginSusie", "dp->d_name = %s\n", dp->d_name);
		
		    if (spEqSuffix(dp->d_name, SP_SUSIE_SUFFIX)) {
			spi = spCreateSusieSPI(dp->d_name);
		    } else {
			spDebug(80, "spInitPluginSusie", "not susie plugin: %s\n", dp->d_name);
		    }
		}
		spCloseDir(dirp);
	    }
	    
	    SetCurrentDirectory(current_dir);
	}
    }
    /*spSetDebugLevel(10);*/
    
    spDebug(80, "spInitPluginSusie", "done\n");

    if (sp_susie_first_spi != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spFreePluginSusie(void)
{
    spFreeSusieFileTypeList();
    spDestroyAllSusieSPI();

    return SP_TRUE;
}

static spSusieSPI spFindSuitableSusieSPI(const char *filename, int typeIndex, FILE **pfp)
{
    spSusieSPI spi = NULL;
    FILE *fp;
    int ret;
    unsigned char header[SP_SUSIE_HEADER_BUFFER_SIZE];
    int size = SP_SUSIE_HEADER_BUFFER_SIZE;
    spSusiePictureInfo picinfo;

    spDebug(80, "spFindSuitableSusieSPI", "filename = %s\n", filename);
    
    if ((fp = spOpenFile(filename, "rb")) != NULL) {
	memset(header, 0, sizeof(header));
	if ((size = (int)fread(header, 1, size, fp)) > 0) {
	    spi = sp_susie_first_spi;
	    while (spi != NULL) {
		spDebug(100, "spFindSuitableSusieSPI", "spi->name = %s, typeIndex = %d, spi->typeOffset = %d, spi->typeCount = %d\n",
			spi->name, typeIndex, spi->typeOffset, spi->typeCount);
		if (typeIndex < 0
		    || (typeIndex >= spi->typeOffset && typeIndex < spi->typeOffset + spi->typeCount)) {
		    if (spi->isSupported(filename, header)) {
			spDebug(60, "spFindSuitableSusieSPI", "**** Suppored by %s ****\n", spi->name);
			memset(&picinfo, 0, sizeof(spSusiePictureInfo));
			ret = (*spi->getPictureInfo)(filename, 0, 0, &picinfo);
			if (picinfo.hInfo != NULL) {
			    LocalFree(picinfo.hInfo);
			}
			
			if (ret == 0 && picinfo.width > 0 && picinfo.height > 0) {
			    spDebug(60, "spFindSuitableSusieSPI", "%s: support %s\n", spi->name, filename);
			    break;
			} else {
			    spDebug(60, "spFindSuitableSusieSPI", "%s: getPictureInfo failed: ret = %d, width = %ld, height = %ld, colorDepth = %d\n",
				    spi->name, ret, picinfo.width, picinfo.height, picinfo.colorDepth);
			}
		    } else {
			spDebug(10, "spFindSuitableSusieSPI", "not supported by %s\n", spi->name);
		    }
		}

		spi = spi->next;
	    }
	} else {
	    spDebug(10, "spFindSuitableSusieSPI", "fread error: size = %d\n", size);
	}
	
	if (pfp != NULL) {
	    *pfp = fp;
	} else {
	    spCloseFile(fp);
	}
    }

    spDebug(80, "spFindSuitableSusieSPI", "spi = %ld\n", (long)spi);
    
    return spi;
}

static spBool spIsSupportedByPluginSusie(const char *filename)
{
    spSusieSPI spi = NULL;
    spBool supported = SP_FALSE;

    spDebug(100, "spIsSupportedByPluginSusie", "in\n");
    
    if (spGetBaseName(filename) == filename && filename[0] == '.') {
    } else {
	if ((spi = spFindSuitableSusieSPI(filename, -1, NULL)) != NULL) {
	    supported = SP_TRUE;
	}
    }
    
    spDebug(100, "spIsSupportedByPluginSusie", "done: supported = %d\n", supported);
    
    return supported;
}

static void *spInitPluginInstanceSusie(const char *lang)
{
    spSusiePluginInstance pinstance;
    
    pinstance = xalloc(1, struct _spSusiePluginInstance);
    memset(pinstance, 0, sizeof(struct _spSusiePluginInstance));
    pinstance->typeIndex = -1;
    
    return (void *)pinstance;
}

static spBool spFreePluginInstanceSusie(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeSusie(void *instance, int index)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;

    if (index == 0 && index < sp_susie_num_file_type_list) {
	pinstance->typeIndex = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static int spGetPluginFileTypeSusie(void *instance)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;
    
    return pinstance->typeIndex;
}

static const char *spGetPluginBestSuffixSusie(void *instance)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;

    if (sp_susie_file_filter_list == NULL) return NULL;
    
    if (pinstance->typeIndex == 0 && pinstance->typeIndex < sp_susie_num_file_type_list) {
	const char *filter;
	const char *suffix;

	filter = sp_susie_file_filter_list[pinstance->typeIndex];
	suffix = spStrChr(filter, '.');
	if (suffix != NULL) {
	    return suffix;
	}
    }
    
    return NULL;
}

static unsigned long spGetPluginDIBCapsSusie(void *instance)
{
    return SP_DI_BITMAP_CAPS_DIRECT_ACCESS | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3 | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4
	| SP_DI_BITMAP_CAPS_READ;
}

static spBool spClosePluginSusie(void *instance)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;

    if (pinstance->filename != NULL) {
	xfree(pinstance->filename);
	pinstance->filename = NULL;
    }
    pinstance->spi = NULL;
    
    return SP_TRUE;
}

static spPluginError spOpenPluginSusie(void *instance, const char *filename, const char *mode)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;

    if (mode[0] == 'w') {
	return SP_PLUGIN_ERROR_OPEN;
    }
    
    if ((pinstance->spi = spFindSuitableSusieSPI(filename, pinstance->typeIndex, NULL)) == NULL) {
	spDebug(10, "spOpenPluginSusie", "can't find suitable Susie plug-in\n");
	return SP_PLUGIN_ERROR_OPEN;
    }
    pinstance->filename = strclone(filename);
	
    spDebug(80, "spOpenPluginSusie", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spGetPluginDIBInfoSusie(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned long *strict_mask)
{
    int ret;
    spSusiePictureInfo picinfo;
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;

    spDebug(80, "spGetPluginDIBInfoSusie", "in\n");
    
    if (pinstance->filename == NULL || pinstance->spi == NULL) {
	return SP_FALSE;
    }

    memset(&picinfo, 0, sizeof(spSusiePictureInfo));
    ret = (*pinstance->spi->getPictureInfo)(pinstance->filename, 0, 0, &picinfo);
    if (picinfo.hInfo != NULL) {
        LocalFree(picinfo.hInfo);
    }

    if (ret != 0 || picinfo.width <= 0 || picinfo.height <= 0) {
	spDebug(80, "spGetPluginDIBInfoSusie", "getPictureInfo failed: %d, colorDepth = %d\n",
		ret, picinfo.colorDepth);
	return SP_FALSE;
    }

    spDebug(80, "spGetPluginDIBInfoSusie", "colorDepth = %d, x_density = %d, y_density = %d, width = %ld, height = %ld\n",
	    picinfo.colorDepth, picinfo.x_density, picinfo.y_density, picinfo.width, picinfo.height);
    
    if (((*strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_SINGLE_RGB_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_FLOAT_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	    && (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && !(dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
	    && dibinfo->upside_down == SP_FALSE)
	|| ((*strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && dibinfo->bit_size != 8)) {
	spDebug(80, "spGetPluginDIBInfoSusie", "unsupported strict mask exists\n");
	return SP_FALSE;
    }

    if (picinfo.colorDepth <= 8) {
	dibinfo->pixel_stride = 3;
    } else {
	dibinfo->pixel_stride = (picinfo.colorDepth + 7) / 8;
    }
    
    if (dibinfo->pixel_stride == 4) {
	dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
    } else {
	dibinfo->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
    }

    dibinfo->width = picinfo.width;
    dibinfo->height = picinfo.height;
    dibinfo->line_stride = 4 * ((dibinfo->width * (dibinfo->pixel_stride * 8) + 31) / 32);
    dibinfo->upside_down = SP_TRUE;
    spDebug(80, "spGetPluginDIBInfoSusie", "pixel_stride = %ld, line_stride = %ld\n",
	    dibinfo->pixel_stride, dibinfo->line_stride);
    
    *strict_mask |= SP_DI_BITMAP_STRICT_ALL;
    
    spDebug(80, "spGetPluginDIBInfoSusie", "done\n");
    
    return SP_TRUE;
}

static int __stdcall spSusieProgressFunc(int nNum, int nDenom, LONG_PTR lData)
{
    spSusiePluginInstance pinstance = (spSusiePluginInstance)lData;
    
    spDebug(80, "spSusieProgressFunc", "nNum = %d, nDenom = %d\n", nNum, nDenom);
    
    return 0;
}

static spBool spReadPluginDIBDataSusie(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    int ret;
    spBool flag;
    long size;
    HANDLE hbmpinfo, hbmpdata;
    BITMAPINFOHEADER *pbmpih;
    void *bmpdata;
    long bmpsize;
    spSusiePluginInstance pinstance = (spSusiePluginInstance)instance;
    
    spDebug(80, "spReadPluginDIBDataSusie", "in\n");

    if (pinstance->filename == NULL || pinstance->spi == NULL) {
	return SP_FALSE;
    }

    ret = (*pinstance->spi->getPicture)(pinstance->filename, 0, 0, &hbmpinfo, &hbmpdata,
					spSusieProgressFunc, (LONG_PTR)pinstance);
    if (ret != 0) {
	spDebug(10, "spReadPluginDIBDataSusie", "GetPicture failed: ret = %d\n", ret);
	return SP_FALSE;
    }
    
    spDebug(80, "spReadPluginDIBDataSusie", "pixel_stride = %ld, line_stride = %ld\n",
	    dibinfo->pixel_stride, dibinfo->line_stride);
    
    flag = SP_FALSE;

    size = dibinfo->height * dibinfo->line_stride;
    pbmpih = (BITMAPINFOHEADER *)LocalLock(hbmpinfo);
    spDebug(80, "spReadPluginDIBDataSusie", "biWidth = %d, biHeight = %d, biPlanes = %d\n",
	    pbmpih->biWidth, pbmpih->biHeight, pbmpih->biPlanes);
    spDebug(80, "spReadPluginDIBDataSusie",
	    "biBitCount = %d, biCompression = %ld (BI_RGB = %ld), biSizeImage = %ld, size = %ld, biXPelsPerMeter = %ld, biYPelsPerMeter = %ld\n",
	    pbmpih->biBitCount, pbmpih->biCompression, BI_RGB, pbmpih->biSizeImage, size,
	    pbmpih->biXPelsPerMeter, pbmpih->biYPelsPerMeter);
    if (pbmpih->biCompression == BI_RGB && pbmpih->biPlanes == 1 
	&& pbmpih->biWidth == dibinfo->width && pbmpih->biHeight == dibinfo->height) {
	flag = SP_TRUE;
	bmpdata = LocalLock(hbmpdata);
	if (pbmpih->biSizeImage == 0) {
	    bmpsize = (long)LocalSize(hbmpdata);
	    spDebug(10, "spReadPluginDIBDataSusie", "bmpsize = %ld, size = %ld\n", bmpsize, size);
	    if (bmpsize > 0) {
		flag = SP_TRUE;
	    }
	} else {
	    bmpsize = pbmpih->biSizeImage;
	}
	
	if (flag) {
	    long i, j;
	    unsigned char r, g, b;
	    unsigned char *ptr;
	    unsigned char *prev_ptr;
	    long zero_fill_size;
	    
	    if (pbmpih->biBitCount == 1 || pbmpih->biBitCount == 4 || pbmpih->biBitCount == 8
		|| pbmpih->biClrUsed > 0) {
		RGBQUAD *prgbquad;
		DWORD *pdwdata;
		long stride;
		long num_palette;
		long palette_index;
		long k;
		long pdwindex;
		long nblock;
		long block_shift;
		long mask_unit = 0L;

		num_palette = (pbmpih->biClrUsed == 0 ? 1L << pbmpih->biBitCount : pbmpih->biClrUsed);
		spDebug(80, "spReadPluginDIBDataSusie", "num_palette = %ld, biClrUsed = %ld\n", num_palette,
			pbmpih->biClrUsed);

		stride = 4 * ((dibinfo->width * pbmpih->biBitCount + 31) / 32);
		prgbquad = (RGBQUAD *)(pbmpih + 1);
		pdwdata = (DWORD *)bmpdata;
		nblock = 32L / pbmpih->biBitCount;
		mask_unit = POW2(pbmpih->biBitCount) - 1L;
		spDebug(80, "spReadPluginDIBDataSusie", "line_stride = %ld, stride = %ld, nblock = %ld, mask_unit = %lx\n",
			dibinfo->line_stride, stride, nblock, mask_unit);
		
		zero_fill_size = dibinfo->line_stride - dibinfo->pixel_stride * dibinfo->width;
		spDebug(80, "spReadPluginDIBDataSusie", "zero_fill_size = %ld\n", zero_fill_size);

		ptr = data;

		for (j = 0; j < dibinfo->height; j++) {
		    prev_ptr = ptr;
		    for (i = 0; i < dibinfo->width; i++) {
			if (pbmpih->biBitCount <= 16) {
			    pdwindex = i / nblock;
			    k = i % nblock;
			    block_shift = pbmpih->biBitCount * k;
			    palette_index = mask_unit & (pdwdata[pdwindex] >> block_shift);
			} else {
			    palette_index = pdwdata[i];
			}

			*ptr++ = prgbquad[palette_index].rgbBlue;
			*ptr++ = prgbquad[palette_index].rgbGreen;
			*ptr++ = prgbquad[palette_index].rgbRed;
			if (dibinfo->pixel_stride == 4) {
			    *ptr++ = 0;
			}
		    }

		    if (zero_fill_size > 0) {
			memset(ptr, 0, zero_fill_size);
			ptr += zero_fill_size;
		    }
		    
		    pdwdata = (DWORD *)((unsigned char *)pdwdata + stride);
		}
	    } else if (pbmpih->biBitCount == 16) {
		long pwstride;
		WORD *pwdata;
		double weight;

		pwstride = 4 * ((dibinfo->width * 16 + 31) / 32);
		pwdata = (WORD *)bmpdata;
		ptr = data;
		
		zero_fill_size = dibinfo->line_stride - dibinfo->pixel_stride * dibinfo->width;
		spDebug(80, "spReadPluginDIBDataSusie", "zero_fill_size = %ld\n", zero_fill_size);
		
		weight = 255.0 / 31.0;
		for (j = 0; j < dibinfo->height; j++) {
		    prev_ptr = ptr;
		    for (i = 0; i < dibinfo->width; i++) {
			r = (unsigned char)(weight * ((0x7c00 & pwdata[i]) >> 10) + 0.5);
			g = (unsigned char)(weight * ((0x3e0 & pwdata[i]) >> 5) + 0.5);
			b = (unsigned char)(weight * (0x1f & pwdata[i]) + 0.5);
			
			*ptr++ = b;
			*ptr++ = g;
			*ptr++ = r;
			if (dibinfo->pixel_stride == 4) {
			    *ptr++ = 0;
			}
		    }

		    if (zero_fill_size > 0) {
			memset(ptr, 0, zero_fill_size);
			ptr += zero_fill_size;
		    }
		    
		    pwdata = (WORD *)((unsigned char *)pwdata + pwstride);
		}
	    } else {
		memcpy(data, bmpdata, MIN(size, bmpsize));
	    }
	}
	LocalUnlock(hbmpdata);
    } else {
	spDebug(10, "spReadPluginDIBDataSusie", "unsupported format\n");
    }
    LocalUnlock(hbmpinfo);
    LocalFree(hbmpdata);
    LocalFree(hbmpinfo);
    
    return flag;
}

static spDIBitmapPluginRec sp_dibitmap_plugin_susie = {
    NULL,
    NULL,

    SP_PLUGIN_DIBITMAP,
    "susie",
    1,
    SP_PLUGIN_PRIORITY_LOWER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginSusie,
    spFreePluginSusie,
    "Susie Input Plugin",
    "Susie Input Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceSusie,
    spFreePluginInstanceSusie,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    NULL,
    NULL,
    NULL,
    spIsSupportedByPluginSusie,
    spSetPluginFileTypeSusie,  
    spGetPluginFileTypeSusie,
    0L,
    NULL,
    NULL,
    spGetPluginBestSuffixSusie,
    
    spGetPluginDIBCapsSusie,
	
    spOpenPluginSusie,              
    spClosePluginSusie,
	
    NULL,
    NULL,
    NULL,
    spGetPluginDIBInfoSusie,
    NULL,   
    NULL,
	
    spReadPluginDIBDataSusie,          
    NULL,
	
    NULL,
    NULL,
};

static spDIBitmapPluginRec *spGetSusieDIBitmapPluginRec(void)
{
    return &sp_dibitmap_plugin_susie;
}

#if 1
spBool spAddSusieDIBitmapPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_dibitmap_plugin_susie, "susie");
}

spBool spRemoveSusieDIBitmapPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_dibitmap_plugin_susie);
}
#else
#include <sp/spPluginMain.h>

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_dibitmap_plugin_susie;
}
#endif
