/*
 *        spCanvas_Win.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>
#include <sp/spSliderP.h>
#include <sp/spImageP.h>
#include <sp/spCanvasP.h>

extern spTopLevel sp_toplevel;

spBool spIsCanvasExposedArch(spComponent component)
{
    if (/*SpComponentPart(component).client_width >= 0
          && SpComponentPart(component).client_height >= 0*/SpCanvasPart(component).exposed) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

void spExposeCanvasCBArch(spComponent component)
{
    return;
}

static void resizeCanvasCB(spComponent component, void *data)
{
    if (SpPrimitiveArch(SpGetWindow(component)).realize_flag == SP_TRUE) {
        spResizeCanvasCB(component, data);
    }
    return;
}

spBool spUpdateCanvasScrollBarArch(spComponent component)
{
    SCROLLINFO sinfo;
    spBool flag = SP_FALSE;
    
    if (SpCanvasPart(component).h_scroll == SP_TRUE) {
        SpCanvasPart(component).h_scroll_coef = spGetScrollCoefWin(SpCanvasPart(component).h_maximum);

        sinfo.cbSize = sizeof(SCROLLINFO);
        sinfo.fMask = SIF_POS | SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
        sinfo.nMin = SpCanvasPart(component).h_minimum / SpCanvasPart(component).h_scroll_coef;
        sinfo.nMax = SpCanvasPart(component).h_maximum / SpCanvasPart(component).h_scroll_coef;
        sinfo.nPage = SpCanvasPart(component).h_page_size / SpCanvasPart(component).h_scroll_coef + 1;
        sinfo.nPos = SpCanvasPart(component).h_value / SpCanvasPart(component).h_scroll_coef;
        sinfo.nTrackPos = 0;
        
        SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_HORZ, &sinfo, TRUE);
        flag = SP_TRUE;
    }
    if (SpCanvasPart(component).v_scroll == SP_TRUE) {
        SpCanvasPart(component).v_scroll_coef = spGetScrollCoefWin(SpCanvasPart(component).v_maximum);

        sinfo.cbSize = sizeof(SCROLLINFO);
        sinfo.fMask = SIF_POS | SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
        sinfo.nMin = SpCanvasPart(component).v_minimum / SpCanvasPart(component).v_scroll_coef;
        sinfo.nMax = SpCanvasPart(component).v_maximum / SpCanvasPart(component).v_scroll_coef;
        sinfo.nPage = SpCanvasPart(component).v_page_size / SpCanvasPart(component).v_scroll_coef + 1;
        sinfo.nPos = SpCanvasPart(component).v_value / SpCanvasPart(component).v_scroll_coef;
        sinfo.nTrackPos = 0;
        
        SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_VERT, &sinfo, TRUE);
        flag = SP_TRUE;
    }

    return flag;
}

void spSetCanvasHScrollBarValueWin(spComponent component)
{
    SCROLLINFO sinfo;

    if (spIsSensitive(component) == SP_FALSE) return;

    spUpdateScrollValueInfoWin(&sinfo, &SpCanvasPart(component).h_value,
                               SpCanvasPart(component).h_minimum, SpCanvasPart(component).h_maximum,
                               SpCanvasPart(component).h_page_size, SpCanvasPart(component).h_scroll_coef);
        
    spDebug(50, "spSetCanvasHScrollBarValueWin","min = %d, max = %d, value = %d\n",
            SpCanvasPart(component).h_minimum, SpCanvasPart(component).h_maximum,
            SpCanvasPart(component).h_value);
    SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_HORZ, &sinfo, 1);
    
    return;
}

void spSetCanvasVScrollBarValueWin(spComponent component)
{
    SCROLLINFO sinfo;

    if (spIsSensitive(component) == SP_FALSE) return;

    spUpdateScrollValueInfoWin(&sinfo, &SpCanvasPart(component).v_value,
                               SpCanvasPart(component).v_minimum, SpCanvasPart(component).v_maximum,
                               SpCanvasPart(component).v_page_size, SpCanvasPart(component).v_scroll_coef);
        
    spDebug(50, "spSetCanvasHScrollBarValueWin","min = %d, max = %d, value = %d\n",
            SpCanvasPart(component).v_minimum, SpCanvasPart(component).v_maximum,
            SpCanvasPart(component).v_value);
    SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_VERT, &sinfo, 1);
    
    return;
}

void spCanvasCreateArch(spComponent component)
{
    spDebug(80, "spCanvasCreateArch", "SpParentPrimitiveArch(component).hwnd = %ld, SpComponentPart(component).component_id = %ld\n",
            (long)SpParentPrimitiveArch(component).hwnd, SpComponentPart(component).component_id);
    
    SpPrimitiveArch(component).hwnd =
        CreateWindowEx((SpCanvasPart(component).border_on == SP_TRUE ? WS_EX_CLIENTEDGE : 0),
                       SP_CANVAS_CLASS_NAME,
                       (!strnone(SpGetName(component)) ? SpGetName(component) : ""),
                       WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN
                       | (SpCanvasPart(component).focusable == SP_TRUE ? WS_TABSTOP : 0)
                       | (SpCanvasPart(component).h_scroll == SP_TRUE ? WS_HSCROLL : 0)
                       | (SpCanvasPart(component).v_scroll == SP_TRUE ? WS_VSCROLL : 0),
                       SpComponentPart(component).x, SpComponentPart(component).y,
                       SpComponentPart(component).current_width,
                       SpComponentPart(component).current_height,
                       SpParentPrimitiveArch(component).hwnd,
                       (HMENU)SpComponentPart(component).component_id,
                       SpTopLevelArch(sp_toplevel).hThisInst,
                       NULL);

    spDebug(100, "spCanvasCreateArch", "call spUpdateCanvasScrollBarArch\n");
    spUpdateCanvasScrollBarArch(component);
    
    spDebug(100, "spCanvasCreateArch", "call spAddCallback\n");
    spAddCallback(component, SP_RESIZE_CALLBACK, resizeCanvasCB, NULL);

    spDebug(100, "spCanvasCreateArch", "done\n");
    
    return;
}

void spCanvasSetParamsArch(spComponent component)
{
    spUpdateCanvasScrollBarArch(component);
    
    return;
}
    
void spCanvasDestroyArch(spComponent component)
{
    return;
}
    
spBool spRefreshCanvasArch(spComponent component)
{
    PAINTSTRUCT paintstruct;
    HDC hdc;

    spDebug(40, "spRefreshCanvasArch", "message = %x, width = %d, height = %d\n", 
            SpPrimitiveArch(component).message,
            SpComponentPart(component).client_width, SpComponentPart(component).client_height);
    
    if (SpPrimitiveArch(component).message == WM_PAINT) {
        spDebug(80, "spRefreshCanvasArch", "WM_PAINT message\n");
        if (GetUpdateRect(SpPrimitiveArch(component).hwnd, NULL, FALSE)) {
            spDebug(80, "spRefreshCanvasArch", "GetUpdateRect succeeded\n");
            spBeginPaintWin(component, &paintstruct);
            hdc = paintstruct.hdc;
        
            spDebug(40, "spRefreshCanvasArch", "erase = %d, left = %d, right = %d, top = %d, bottom = %d\n",
                    paintstruct.fErase, paintstruct.rcPaint.left, paintstruct.rcPaint.right,
                    paintstruct.rcPaint.top, paintstruct.rcPaint.bottom); 
            BitBlt(hdc,
                   paintstruct.rcPaint.left, paintstruct.rcPaint.top,
                   paintstruct.rcPaint.right - paintstruct.rcPaint.left,
                   paintstruct.rcPaint.bottom - paintstruct.rcPaint.top,
                   SpPrimitiveArch(component).memdc, 
                   paintstruct.rcPaint.left, paintstruct.rcPaint.top,
                   SRCCOPY);

            spEndPaintWin(component, &paintstruct);
        } else {
            spDebug(10, "spRefreshCanvasArch", "GetUpdateRect failed\n");
            /* to prevent from getting GetUpdateRect failed */
            InvalidateRect(SpPrimitiveArch(component).hwnd, NULL, FALSE);
        }
    } else {
        hdc = GetDC(SpPrimitiveArch(component).hwnd);
        BitBlt(hdc, 0, 0,
               SpComponentPart(component).client_width, SpComponentPart(component).client_height,
               SpPrimitiveArch(component).memdc, 0, 0, SRCCOPY);
        ReleaseDC(SpPrimitiveArch(component).hwnd, hdc);
        ValidateRect(SpPrimitiveArch(component).hwnd, NULL);
        spDebug(80, "spRefreshCanvasArch", "not WM_PAINT message, BitBlt done\n");
    }
    
    spDebug(80, "spRefreshCanvas", "done\n");
    
    return SP_TRUE;
}

spBool spRedrawCanvasArch(spComponent component)
{
    spRedrawImageArch(component);
    return SP_TRUE;
}

spBool spSetCanvasCaptureArch(spComponent window)
{
    return SP_TRUE;
}

spBool spSetCanvasCursorArch(spComponent component, spCursor cursor)
{
    SpPrimitiveArch(component).hcursor = SpCursorArch(cursor).hcursor;
    
    return SP_TRUE;
}

spBool spUnsetCanvasCursorArch(spComponent component)
{
    SpPrimitiveArch(component).hcursor = NULL;
    
    return SP_TRUE;
}

spBool spIsCanvasFocusedArch(spComponent component)
{
    if (SpPrimitiveArch(component).hwnd == GetFocus()) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spZoomScrollCanvasArch(spComponent component)
{
    return SP_TRUE;
}

spCallback *spFindDrawCallbackArch(spComponent component, long *index, spBool resize_flag)
{
    long k;
    spComponent window;
    spCallback *callback;

    window = SpGetWindow(component);
    if (spIsPrimitive(window) == SP_FALSE) {
        return NULL;
    }

    for (k = *index; k < SpPrimitivePart(window).num_callback; k++) {
        callback = &SpPrimitivePart(window).callbacks[k];
        if (callback->call_func == resizeCanvasCB) {
            continue;
        }
        if ((resize_flag == SP_TRUE && callback->component == component && callback->message == WM_SIZE)
            || (resize_flag == SP_FALSE && callback->component == component && callback->message == WM_PAINT)) {
            *index = k;
            return callback;
        }
    }

    return NULL;
}
