#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

typedef struct _spDIBitmapMac *spDIBitmapMac;
struct _spDIBitmapMac {
    GWorldPtr gworld;
    PixMapHandle pixmap;
    Ptr base_addr;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
};

void *spGetDIBitmapGWorldMac(spDIBitmap dibitmap)
{
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)dibitmap->dibitmap;
    
    return dibitmapmac->gworld;
}

spBool spCheckBasicDIBitmapStrictMaskMac(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    spDebug(100, "spCheckBasicDIBitmapStrictMaskMac",
            "pixel_format = %lx, upside_down = %d, bit_size = %ld, pixel_stride = %ld, line_stride = %ld\n",
            info->pixel_format, info->upside_down, info->bit_size, info->pixel_stride, info->line_stride);
    
    if (((strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
	    && !(info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
	    && info->upside_down != SP_FALSE)
	|| ((strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && info->bit_size != 8)
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	    && info->pixel_stride != 4)
	|| ((strict_mask & SP_DI_BITMAP_STRICT_LINE_STRIDE)
	    && (info->line_stride != 0 && info->line_stride != info->width * info->pixel_stride))) {
        spDebug(80, "spCheckBasicDIBitmapStrictMaskMac", "not supported: strict_mask = %lx\n", strict_mask);
	return SP_FALSE;
    } else {
	return SP_TRUE;
    }
}

void *spCreateDIBitmapMac(spDIBitmapInfoInternal *info, GWorldPtr gworld)
{
    spDIBitmapMac dibitmapmac;

    if (gworld == nil) return SP_FALSE;

    dibitmapmac = xalloc(1, struct _spDIBitmapMac);
    dibitmapmac->gworld = gworld;
    dibitmapmac->pixmap = NULL;
    dibitmapmac->base_addr = NULL;
	
    info->pixel_format = SP_DI_PIXEL_FORMAT_ARGB;
    info->upside_down = SP_FALSE;
    info->bit_size = 8;
    info->pixel_stride = 4;
    info->line_stride = info->width * info->pixel_stride;
    info->data = NULL;
	
    return dibitmapmac;
}

spBool spUpdateDIBitmapInfoArch(spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    if (spCheckBasicDIBitmapStrictMaskMac(info, *strict_mask) == SP_FALSE) {
        spDebug(10, "spUpdateDIBitmapInfoArch", "spCheckBasicDIBitmapStrictMaskMac failed\n");
	return SP_FALSE;
    }
    
    info->pixel_format = SP_DI_PIXEL_FORMAT_ARGB;
    info->upside_down = SP_FALSE;
    info->bit_size = 8;
    info->pixel_stride = 4;
    info->line_stride = info->width * info->pixel_stride;
    
    *strict_mask |= (SP_DI_BITMAP_STRICT_ALL & ~SP_DI_BITMAP_STRICT_ALPHA_STATE);
    
    return SP_TRUE;
}

void *spCreateDIBitmapArch(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    Rect rect;
    GWorldPtr gworld;
    GWorldFlags flags = 0;

    SetRect(&rect, 0, 0, info->width, info->height);
	
    if (noErr == NewGWorld(&gworld, 32, &rect, NULL, NULL, flags)) {
	return spCreateDIBitmapMac(info, gworld);
    } else {
	return NULL;
    }
}

spBool spDestroyDIBitmapArch(void *bitmap)
{
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;
	
    if (dibitmapmac->gworld != NULL) {
	DisposeGWorld(dibitmapmac->gworld);
    }
    xfree(dibitmapmac);

    return SP_TRUE;
}


spBool spLockDIBitmapArch(spDIBitmapInfoInternal *info, void *bitmap)
{
    PixMapHandle pixmap;
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;
	
    if (dibitmapmac->gworld == NULL) return SP_FALSE;
	
#if TARGET_API_MAC_CARBON
    LockPortBits(dibitmapmac->gworld);
#endif
    if ((pixmap = GetGWorldPixMap(dibitmapmac->gworld)) != NULL) {
	if (GetPixBaseAddr(pixmap) == NULL || LockPixels(pixmap) == false) {
	    Rect rect;
		
	    SetRect(&rect, 0, 0, info->width, info->height);
		
	    if (UpdateGWorld(&dibitmapmac->gworld,
			     0, &rect, NULL, NULL, clipPix) == gwFlagErr) {
		pixmap = NULL;
	    } else if (LockPixels(pixmap) == false) {
		pixmap = NULL;
	    }
	}
	if (pixmap != NULL) {
	    dibitmapmac->pixmap = pixmap;
	    dibitmapmac->base_addr = GetPixBaseAddr(pixmap);
	    info->line_stride = GetPixRowBytes(pixmap);
	    info->data = (unsigned char *)dibitmapmac->base_addr;
	    spDebug(50, "spLockDIBitmap", "base_addr = %ld, line_stride = %ld, rowBytes = %ld\n",
		    dibitmapmac->base_addr, info->line_stride, (long)GetPixRowBytes(pixmap));

	    GetGWorld(&dibitmapmac->save_gwp, &dibitmapmac->save_gdh);
	    SetGWorld(dibitmapmac->gworld, NULL);

	    return SP_TRUE;
	}
    }
#if TARGET_API_MAC_CARBON
    UnlockPortBits(dibitmapmac->gworld);
#endif
    return SP_FALSE;
}

spBool spUnlockDIBitmapArch(spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;

    if (dibitmapmac->gworld == NULL) return SP_FALSE;
	
    if (dibitmapmac->pixmap != NULL) {
	SetGWorld(dibitmapmac->save_gwp, dibitmapmac->save_gdh);
	    
	spDebug(50, "spUnlockDIBitmap", "base_addr = %ld\n", dibitmapmac->base_addr);
	    
	UnlockPixels(dibitmapmac->pixmap);
	dibitmapmac->pixmap = NULL;
	dibitmapmac->base_addr = NULL;
	info->data = NULL;
	    
#if TARGET_API_MAC_CARBON
	UnlockPortBits(dibitmapmac->gworld);
	if (QDIsPortBuffered(dibitmapmac->gworld)) {
	    QDFlushPortBuffer(dibitmapmac->gworld, NULL);
	}
#endif
    }

    return SP_TRUE;
}

spBool spStretchDIBitmapToImageMac(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest,
				   int src_x, int src_y, int src_width, int src_height,
				   int dest_x, int dest_y, int dest_width, int dest_height)
{
    spBool flag = SP_FALSE;
    Rect src_rect;
    Rect dest_rect;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    PixMapHandle dest_pixmap;
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;

    if ((dest_pixmap = spGetLockedPixmapMac(dest)) != nil) {
	spDebug(80, "spStretchDIBitmapToImageMac",
		"src_width = %d, src_height = %d, dest_width = %d, dest_height = %d\n",
		src_width, src_height, dest_width, dest_height);
	
	SetRect(&src_rect, src_x, src_y, src_x + src_width, src_y + src_height);
	SetRect(&dest_rect, dest_x, dest_y, dest_x + dest_width, dest_y + dest_height);
	    
	spGetOriginalGWorldMac(&save_gwp, &save_gdh);
	SetGWorld(SpPrimitiveArch(dest).gworld, NULL);

	/* set normal RGB */
	spGetOriginalRGBMac();
	spSetNormalRGBMac();
	    
	/* copy pixmap */
	CopyBits((BitMap *)*(dibitmapmac->pixmap), (BitMap *)*dest_pixmap,
		 &src_rect, &dest_rect, srcCopy, NULL);
	    
	/* restore RGB */
	spSetOriginalRGBMac();
	    
	SetGWorld(save_gwp, save_gdh);
	    
	spUnlockPixmapMac(dest, dest_pixmap);
	
	flag = SP_TRUE;
    }
    
    return flag;
}

spBool spCopyDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest, int src_x, int src_y,
				 int width, int height, int dest_x, int dest_y)
{
    return spStretchDIBitmapToImageMac(info, bitmap, dest,
				       src_x, src_y, width, height,
				       dest_x, dest_y, width, height);
}

spBool spStretchDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest,
				    int src_x, int src_y, int src_width, int src_height,
				    int dest_x, int dest_y, int dest_width, int dest_height)
{
    return spStretchDIBitmapToImageMac(info, bitmap, dest,
				       src_x, src_y, src_width, src_height,
				       dest_x, dest_y, dest_width, dest_height);
}

spBool spStretchImageToDIBitmapMac(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, 
				   int src_x, int src_y, int src_width, int src_height,
				   int dest_x, int dest_y, int dest_width, int dest_height)
{
    spBool flag = SP_FALSE;
    Rect src_rect;
    Rect dest_rect;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    PixMapHandle src_pixmap;
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;

    if ((src_pixmap = spGetLockedPixmapMac(src)) != nil) {
	SetRect(&src_rect, src_x, src_y, src_x + src_width, src_y + src_height);
	SetRect(&dest_rect, dest_x, dest_y, dest_x + dest_width, dest_y + dest_height);
	    
	spGetOriginalGWorldMac(&save_gwp, &save_gdh);
	SetGWorld(dibitmapmac->gworld, NULL);

	/* set normal RGB */
	spGetOriginalRGBMac();
	spSetNormalRGBMac();
	    
	/* copy pixmap */
	CopyBits((BitMap *)*src_pixmap, (BitMap *)*(dibitmapmac->pixmap),
		 &src_rect, &dest_rect, srcCopy, NULL);
	    
	/* restore RGB */
	spSetOriginalRGBMac();
	    
	SetGWorld(save_gwp, save_gdh);
	    
	spUnlockPixmapMac(src, src_pixmap);
	
	flag = SP_TRUE;
    }
	
    return flag;
}

void *spCopyImageToDIBitmapArch(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, int src_x, int src_y,
				int width, int height, int dest_x, int dest_y)
{
    if (spStretchImageToDIBitmapMac(src, info, bitmap, 
				    src_x, src_y, width, height,
				    dest_x, dest_y, width, height)) {
	return bitmap;
    } else {
	return NULL;
    }
}

typedef struct _spDIBitmapInstanceArch {
    int type_index;

    char mode[8];
    FSSpec fsspec;
    OSType file_type;
    
    GraphicsImportComponent gic;
    Rect rect;

    GraphicsExportComponent gec;
} *spDIBitmapInstanceArch;

void *spInitDIBitmapInstanceArch(const char *name)
{
    spDIBitmapInstanceArch dibinstance = NULL;
    
    dibinstance = xalloc(1, struct _spDIBitmapInstanceArch);
    memset(dibinstance, 0, sizeof(struct _spDIBitmapInstanceArch));
    
    spDebug(80, "spInitDIBitmapInstanceArch", "done\n");
    
    return dibinstance;
}

spBool spFreeDIBitmapInstanceArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    xfree(dibinstance);
    
    spDebug(80, "spFreeDIBitmapInstanceArch", "done\n");
    
    return SP_TRUE;
}

int spGetDIBitmapPriorityArch(void *instance)
{
    return SP_PLUGIN_PRIORITY_MIDDLE;
}

unsigned long spGetDIBitmapCapsArch(void *instance)
{
    return SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_READ_USE_BITMAP_ARGUMENT | SP_DI_BITMAP_CAPS_WRITE_USE_BITMAP_ARGUMENT
	| SP_DI_BITMAP_CAPS_ALPHA | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4
	| SP_DI_BITMAP_CAPS_COPY_ALPHA | SP_DI_BITMAP_CAPS_COPY_IMAGE | SP_DI_BITMAP_CAPS_COPY_IMAGE_STRETCHABLE;
}

static const char *sp_dibitmap_file_type_list_arch[] = {
    "pict",
    "png",
    "tiff",
    "jpeg",
    "bmp",
    NULL,
};

static const char *sp_dibitmap_file_desc_list_arch[] = {
    "PICT",
    "PNG",
    "TIFF",
    "JPEG",
    "BMP",
    NULL,
};

static const char *sp_dibitmap_file_filter_list_arch[] = {
    "*.pct",
    "*.png",
    "*.tif",
    "*.jpg",
    "*.bmp",
    NULL,
};

const char **spGetDIBitmapFileTypeListArch(void *instance)
{
    return sp_dibitmap_file_type_list_arch;
}

const char **spGetDIBitmapFileDescListArch(void *instance)
{
    return sp_dibitmap_file_desc_list_arch;
}

const char **spGetDIBitmapFileFilterListArch(void *instance)
{
    return sp_dibitmap_file_filter_list_arch;
}

static int getTypeIndexFromSuffix(const char *suffix)
{
    int type_index = -1;
    
    if (strcaseeq(suffix, ".pct") || strcaseeq(suffix, ".pict") || strcaseeq(suffix, ".pic")) {
	type_index = 0;
    } else if (strcaseeq(suffix, ".png")) {
	type_index = 1;
    } else if (strcaseeq(suffix, ".tif") || strcaseeq(suffix, ".tiff")) {
	type_index = 2;
    } else if (strcaseeq(suffix, ".jpg") || strcaseeq(suffix, ".jpeg")) {
	type_index = 3;
    } else if (strcaseeq(suffix, ".bmp")) {
	type_index = 4;
    }

    spDebug(80, "getTypeIndexFromSuffix", "suffix = %s, type_index = %d\n", suffix, type_index);
    
    return type_index;
}

spBool spIsDIBitmapSupportedArch(void *instance, const char *filename)
{
    spBool supported = SP_FALSE;

    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (getTypeIndexFromSuffix(filename) >= 0) {
	    supported = SP_TRUE;
	}
    } else {
	FSSpec fsspec;
	GraphicsImportComponent gic;
	
	if (spPathToFSSpecMac(filename, &fsspec) == SP_TRUE) {
	    if (spOpenPictFromFSSpecMac(&fsspec, &gic, NULL) == SP_TRUE) {
		supported = SP_TRUE;
		CloseComponent(gic);
            } else {
                spDebug(50, "spIsDIBitmapSupportedArch", "spOpenPictFromFSSpecMac failed\n");
	    }
	} else {
	    spDebug(50, "spIsDIBitmapSupportedArch", "spPathToFSSpecMac failed\n");
        }
    }
    
    return supported;
}

spBool spSetDIBitmapFileTypeArch(void *instance, int index)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (index >= 0 && index <= 4) {
	dibinstance->type_index = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

int spGetDIBitmapFileTypeArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    return dibinstance->type_index;
}

const char *spGetDIBitmapBestSuffixArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->type_index == 0) {
	return ".pct";
    } else if (dibinstance->type_index == 1) {
	return ".png";
    } else if (dibinstance->type_index == 2) {
	return ".tif";
    } else if (dibinstance->type_index == 3) {
	return ".jpg";
    } else if (dibinstance->type_index == 4) {
	return ".bmp";
    } else {
	return NULL;
    }
}

spBool spOpenDIBitmapFileArch(void *instance, const char *filename, const char *mode)
{    
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (spPathToFSSpecMac(filename, &dibinstance->fsspec) == SP_FALSE) {
        spDebug(10, "spOpenDIBitmapFileArch", "spPathToFSSpecMac failed\n");
	return SP_FALSE;
    }

    spStrCopy(dibinstance->mode, sizeof(dibinstance->mode), mode);

    if (mode[0] == 'w') {
	if (dibinstance->type_index < 0) {
	    dibinstance->type_index = getTypeIndexFromSuffix(spGetSuffix(filename));
	}
	
	if (dibinstance->type_index == 0) {
	    dibinstance->file_type = kQTFileTypePicture;
	} else if (dibinstance->type_index == 1) {
	    dibinstance->file_type = kQTFileTypePNG;
	} else if (dibinstance->type_index == 2) {
	    dibinstance->file_type = kQTFileTypeTIFF;
	} else if (dibinstance->type_index == 3) {
	    dibinstance->file_type = kQTFileTypeJPEG;
	} else if (dibinstance->type_index == 4) {
	    dibinstance->file_type = kQTFileTypeBMP;
	} else {
	    spDebug(10, "spOpenDIBitmapFileArch", "invalid type_index: %d\n", dibinstance->type_index);
	    return SP_FALSE;
	}
	
	if (OpenADefaultComponent(GraphicsExporterComponentType, dibinstance->file_type, &dibinstance->gec) != noErr) {
	    spDebug(10, "spOpenDIBitmapFileArch", "OpenADefaultComponent failed\n");
	    return SP_FALSE;
	}
    } else {
	dibinstance->file_type = 0;
	if (spOpenPictFromFSSpecMac(&dibinstance->fsspec, &dibinstance->gic, &dibinstance->rect) == SP_FALSE) {
	    spDebug(10, "spOpenDIBitmapFileArch", "spOpenPictFromFSSpecMac failed\n");
	    return SP_FALSE;
	}
	
	if (spGetFileTypeFromGraphicsImportComponentMac(&dibinstance->gic, &dibinstance->file_type) == SP_FALSE) {
	    FInfo finfo;
	    
	    if (FSpGetFInfo(&dibinstance->fsspec, &finfo) == noErr) {
		dibinstance->file_type = finfo.fdType;
	    }
	}

	if (dibinstance->file_type == kQTFileTypePicture) {
	    dibinstance->type_index = 0;
	} else if (dibinstance->file_type == kQTFileTypePNG) {
	    dibinstance->type_index = 1;
	} else if (dibinstance->file_type == kQTFileTypeTIFF) {
	    dibinstance->type_index = 2;
	} else if (dibinstance->file_type == kQTFileTypeJPEG) {
	    dibinstance->type_index = 3;
	} else if (dibinstance->file_type == kQTFileTypeBMP) {
	    dibinstance->type_index = 4;
	} else {
	    spDebug(10, "spOpenDIBitmapFileArch", "invalid file type: %c%c%c%c\n",
		    (char)(dibinstance->file_type>>24),
		    (char)(dibinstance->file_type>>16),
		    (char)(dibinstance->file_type>>8),
		    (char)(dibinstance->file_type));
	    CloseComponent(dibinstance->gic);
	    return SP_FALSE;
	}
    }
    
    spDebug(10, "spOpenDIBitmapFileArch", "file type: %c%c%c%c\n",
	    (char)(dibinstance->file_type>>24),
	    (char)(dibinstance->file_type>>16),
	    (char)(dibinstance->file_type>>8),
	    (char)(dibinstance->file_type));
    
    return SP_TRUE;
}

spBool spCloseDIBitmapFileArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (dibinstance->mode[0] == 'w') {
	CloseComponent(dibinstance->gec);
    } else {
	CloseComponent(dibinstance->gic);
    }
    dibinstance->mode[0] = NUL;
    
    return SP_TRUE;
}

spBool spGetDIBitmapInfoArch(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->mode[0] == 'w') {
	return SP_FALSE;
    }

    if (spUpdateDIBitmapInfoArch(info, strict_mask) == SP_FALSE) {
        spDebug(10, "spGetDIBitmapInfoArch", "spUpdateDIBitmapInfoArch failed\n");
	return SP_FALSE;
    }
    
    info->width = spGetRectWidthMac(dibinstance->rect);
    info->height = spGetRectHeightMac(dibinstance->rect);

    return SP_TRUE;
}

void *spReadDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;

    if (spDrawImportedPictToLockedGWorldMac(dibitmapmac->gworld, &dibinstance->rect, &dibinstance->gic)) {
	return bitmap;
    } else {
        spDebug(10, "spReadDIBitmapArch", "spDrawImportedPictToLockedGWorldMac failed\n");
	return NULL;
    }
}

spBool spWriteDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    spDIBitmapMac dibitmapmac = (spDIBitmapMac)bitmap;

    return spWritePictFromGWorldMac(dibitmapmac->gworld, &dibinstance->gic,
				    &dibinstance->fsspec, dibinstance->file_type);
}

spBool spReverseDIBitmapArch(spDIBitmapInfoInternal *info, void *dibdata)
{
    return spReverseDIBitmapDefault(info, dibdata);
}

spBool spFlipDIBitmapArch(spDIBitmapInfoInternal *info, void *dibdata, spBool lr_flag)
{
    return spFlipDIBitmapDefault(info, dibdata, lr_flag);
}
