/*
 *	spFrame.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spComponent.h>
#include <sp/spTopLevel.h>
#include <sp/spPrimitive.h>
#include <sp/spDialog.h>
#include <sp/spMenu.h>

#include <sp/spFrameP.h>

extern spTopLevel sp_toplevel;

static spParamTable sp_frame_param_tables[] = {
    {SppIconName, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.icon_name), ""},
    {SppTaskTrayIcon, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.task_tray_icon), ""},
    {SppTaskTrayTip, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.task_tray_tip), ""},
    {SppWindowType, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.window_type), SP_NORMAL_WINDOW_STRING},
    {SppPopupStyle, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.popup_style), SP_MODELESS_POPUP_STRING},
    {SppCloseStyle, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.close_style), NULL},
    {SppResizable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.resize_flag), SP_TRUE_STRING},
    {SppIconfiable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.iconfy_flag), SP_TRUE_STRING},
    {SppSimplifiable, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.simplify_flag), SP_FALSE_STRING},
    {SppTaskTrayVisible, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.task_tray_visible), SP_FALSE_STRING},
    {SppHelpButtonVisible, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.help_button_visible), SP_FALSE_STRING},
    {SppInternalWindow, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.internal_flag), SP_FALSE_STRING},
    {SppParentWindow, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.parent_window), NULL},
    {SppSystemWindow, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.system_window), NULL},
    {SppInitialX, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.x), "0"},
    {SppInitialY, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.y), "0"},
    {SppTopMost, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.topmost_flag), SP_FALSE_STRING},
    {SppFrameButtonType, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spFrame, frame.button_type), "-1"},
};

spFrameClassRec SpFrameClassRec = {
    /* spObjectClassPart */
    {
	SpFrame,
	(spObjectClass)&SpPrimitiveClassRec,
	sizeof(spFrameRec),
	spArraySize(sp_frame_param_tables),
	sp_frame_param_tables,
	spFramePartInit,
	spFramePartFree,
	SP_FALSE,
	NULL,
	NULL,
	spFrameCreate,
	spFrameDestroy,
	spFrameSetParams,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_TRUE,
	SP_TRUE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPrimitiveClassPart */
    {
	0,
    },
    /* spFrameClassPart */
    {
	0,
    },
};

spComponentClass SpFrameClass = (spComponentClass)&SpFrameClassRec;

void spFramePartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpComponentPart(component).orientation = SP_VERTICAL;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;
    SpComponentPart(component).visible_flag = SP_FALSE;

    SpFramePart(component).menu_bar = NULL;
    SpFramePart(component).tool_bar = NULL;
    SpFramePart(component).status_bar = NULL;
    SpFramePart(component).task_tray_menu = NULL;
    SpFramePart(component).default_button = NULL;
    SpFramePart(component).cancel_button = NULL;
    SpFramePart(component).parent_window = NULL;
    SpFramePart(component).toplevel = sp_toplevel;
    SpFramePart(component).system_window = NULL;
    SpFramePart(component).timer = NULL;
    
    if (spIsSubClass(component, SpMainFrame) == SP_TRUE) {
	SpFramePart(component).close_style = SP_DESTROY_CLOSE;
    } else {
	SpFramePart(component).close_style = SP_NO_CLOSE;
    }
    
    SpFramePart(component).position_set_flag = SP_FALSE;

    spFramePartInitArch(component);
    
    return;
}

void spFramePartFree(spObject object)
{
    spComponent component = (spComponent)object;
    
    spFramePartFreeArch(component);
    return;
}

spBool spFrameCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spDebug(30, "spFrameCreate", "in\n");

    spDebug(30, "spFrameCreate", "width = %d, height = %d, current_width = %d, current_height = %d\n",
	    SpComponentPart(component).width, SpComponentPart(component).height,
	    SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
	spPrimitiveScaleDefaultSizeArch(component);
    spFrameCreateArch(component);
    
    if (SpComponentPart(component).call_func != NULL) {
	spAddCallback(component, SP_CLOSE_CALLBACK,
		      SpComponentPart(component).call_func,
		      SpComponentPart(component).call_data);
    }

    return SP_TRUE;
}

spBool spFrameDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spDebug(30, "spFrameDestroy", "in\n");

    spFrameDestroyArch(component);
    spPrimitiveDestroyArch(component);

    return SP_TRUE;
}

spBool spFrameSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spDebug(30, "spFrameSetParams", "in\n");
    
    spFrameSetParamsArch(component);

    return SP_TRUE;
}

spBool spIsFrame(spComponent component)
{
    if (spIsSubClass(component, SpFrame) == SP_TRUE
	|| spIsSubClass(component, SpMainFrame) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsMainFrame(spComponent component)
{
    if (spIsSubClass(component, SpMainFrame) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spComponent spCreateFrame(const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsTopLevelCreated() == SP_FALSE) spError(1, "spInitialize should be called.\n");
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    spDebug(60, "spCreateFrame", "spGetArgs done\n");
    
    return spCreateComponentArg(SpFrameClass, SpFrame, name, NULL, args, num_arg);
}

spComponent spCreateMainFrame(const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsTopLevelCreated() == SP_FALSE) spError(1, "spInitialize should be called.\n");
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpFrameClass, SpMainFrame, name, NULL, args, num_arg);
}

spTopLevel spGetTopLevel(spComponent component)
{
    spComponent window;
    
    if (component == NULL) return NULL;

    window = SpGetWindow(component);
    
    if (spIsFrame(window) == SP_FALSE) {
	return NULL;
    } else {
	return SpFramePart(window).toplevel;
    }
}

spBool spIsFrameToolBarVisible(spComponent component)
{
    if (spIsFrame(component) == SP_TRUE
	&& spIsVisible(SpFramePart(component).tool_bar) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsFrameStatusBarVisible(spComponent component)
{
    if (spIsFrame(component) == SP_TRUE
	&& spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spTimerList spAddTimerList(spComponent frame, long interval_ms,
			   spTimerCallbackFunc call_func, void *call_data)
{
    spTimerList prev_timer, timer;

    prev_timer = SpFramePart(frame).timer;
    while (prev_timer != NULL && prev_timer->next != NULL) {
	prev_timer = prev_timer->next;
    }
    
    timer = xalloc(1, struct _spTimerList);
    memset(timer, 0, sizeof(struct _spTimerList));
    if (prev_timer != NULL) {
	timer->prev = prev_timer;
	prev_timer->next = timer;
    }
    timer->frame = frame;
    timer->interval_ms = interval_ms;
    timer->call_func = call_func;
    timer->call_data = call_data;

    if (SpFramePart(frame).timer == NULL) {
	SpFramePart(frame).timer = timer;
    }

    return timer;
}

void spRemoveTimerList(spComponent frame, spTimerList timer)
{
    if (SpFramePart(frame).timer == timer) {
	SpFramePart(frame).timer = timer->next;
    }
    if (timer->prev != NULL) {
	timer->prev->next = timer->next;
    }
    if (timer->next != NULL) {
	timer->next->prev = timer->prev;
    }
    xfree(timer);
    
    return;
}

spTimerList spFindTimerList(spComponent frame, spTimerId timer_id)
{
    spTimerList timer;

    timer = SpFramePart(frame).timer;

    while (timer != NULL) {
	if (timer->timer_id == timer_id) {
	    break;
	}
	timer = timer->next;
    }

    return timer;
}

spTimerId spSetTimer(spComponent frame, long interval_ms,
		     spTimerCallbackFunc call_func, void *call_data)
{
    spTimerList timer;

    if (spIsFrame(frame) == SP_FALSE) {
	frame = spGetWindow(frame);
    }
    
    if (spIsFrame(frame) == SP_FALSE
	|| interval_ms == 0 || call_func == NULL) return SP_TIMER_ID_NONE;

    timer = spAddTimerList(frame, interval_ms, call_func, call_data);
    
    if ((timer->timer_id = spSetTimerArch(timer)) == SP_TIMER_ID_NONE) {
	spRemoveTimerList(frame, timer);
	return SP_TIMER_ID_NONE;
    }

    return timer->timer_id;
}

spBool spKillTimer(spComponent frame, spTimerId timer_id)
{
    spBool flag;
    spTimerList timer;
    
    if (spIsFrame(frame) == SP_FALSE) {
	frame = spGetWindow(frame);
    }
    
    if (spIsFrame(frame) == SP_FALSE
	|| timer_id == SP_TIMER_ID_NONE) return SP_FALSE;

    if ((timer = spFindTimerList(frame, timer_id)) != NULL) {
	flag = spKillTimerArch(timer);
	spDebug(50, "spKillTimer", "spKillTimerArch done: flag = %d\n", flag);
#if 0	/* spRemoveTimerList called in the arch-depend function */
	spRemoveTimerList(frame, timer);
	spDebug(50, "spKillTimer", "spRemoveTimerList done\n");
#endif
    } else {
	flag = SP_FALSE;
    }

    return flag;
}

spBool spAddDropCallback(spComponent frame, spDropCallbackFunc call_func, void *call_data)
{
    if (spIsFrame(frame) == SP_FALSE || call_func == NULL) return SP_FALSE;
    
    return spAddDropCallbackArch(frame, call_func, call_data);
}

spBool spRemoveDropCallback(spComponent frame, spDropCallbackFunc call_func, void *call_data)
{
    if (spIsFrame(frame) == SP_FALSE || call_func == NULL) return SP_FALSE;
    
    return spAddDropCallbackArch(frame, NULL, NULL);
}

void spAddFrameCallback(spComponent frame, spDrawFunc draw_func,
			spMouseFunc mouse_func, spKeyFunc key_func,
			spForegroundFunc foreground_func, spBackgroundFunc background_func, 
			void *data)
{
#if defined(MACOS) && !defined(COCOA)
    if (spIsFrame(frame) == SP_FALSE) return;
    
    spAddFrameCallbackMac(frame, draw_func, mouse_func, key_func,
			  foreground_func, background_func, data);
#endif
    return;
}

spBool spSetTaskTrayMenu(spComponent frame, spComponent menu)
{
    if (spIsFrame(frame) == SP_FALSE || spIsSubClass(menu, SpPopupMenu) == SP_FALSE) return SP_FALSE;

    SpFramePart(frame).task_tray_menu = menu;
    
    return SP_TRUE;
}
