/*
 *	spImage.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spDialog.h>

#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>

spBool spImageCreate(spComponent component)
{
    spDebug(80, "spImageCreate", "in\n");
    
    if (spIsCreated(component) == SP_FALSE) {
	spImageCreateArch(component);
	SpObjectPart(component).create_flag = SP_TRUE;
    }
    if (!strnone(SpComponentPart(component).description)) {
	if (spImageLoadArch(component, SpComponentPart(component).description) == SP_FALSE) {
	    return SP_FALSE;
	}
    }
    
    if (SpComponentPart(component).call_func != NULL) {
	SpPrimitivePart(component).draw_func = SpComponentPart(component).call_func;
	SpPrimitivePart(component).draw_data = SpComponentPart(component).call_data;
    }
    spDebug(80, "spImageCreate", "call spRedrawImage: (%d, %d)\n", SpComponentPart(component).width, SpComponentPart(component).height);
    spRedrawImage(component, SpComponentPart(component).width, SpComponentPart(component).height);

    spDebug(80, "spImageCreate", "done\n");
    
    return SP_TRUE;
}

spBool spIsImage(spComponent component)
{
    return spIsSubClass(component, SpImage);
}

/* without extension case: .bmp (Windows), .xpm (Xm, gtk) */
spComponent spCreateImageFromFile(spComponent parent, const char *filename, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

    if (spIsCreated(parent) == SP_FALSE)
	return NULL;
    
    spDebug(80, "spCreateImageFromFile", "filename = %s\n", filename);
    
    spSetArg(args[num_arg], SppDescription, filename); num_arg++;

    va_start(argp, filename);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    return spCreateComponentArg(SpPrimitiveClass, SpImage, spGetBaseName(filename), parent, args, num_arg);
}

spComponent spCreateImage(spComponent parent, const char *name,
			  int width, int height, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE)
	return NULL;
    
    spSetArg(args[num_arg], SppWidth, width); num_arg++;
    spSetArg(args[num_arg], SppHeight, height); num_arg++;

    va_start(argp, height);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    return spCreateComponentArg(SpPrimitiveClass, SpImage, name, parent, args, num_arg);
}

void spDrawImage(spComponent component)
{
    if (spIsDrawable(component) == SP_FALSE) return;
    
    spDebug(30, "spDrawImage", "in\n");

    if (!(spIsCanvas(component) == SP_TRUE && SpCanvasPart(component).draw_background == SP_FALSE)) {
	if (spIsSubClass(component, SpImage) == SP_FALSE
	    || strnone(SpComponentPart(component).description)) {
            spDebug(30, "spDrawImage", "call spDrawImageArch\n");
	    spDrawImageArch(component);
	}
    }

#if 0
    if (SpPrimitivePart(component).draw_func != NULL) {
	spDebug(30, "spDrawImage", "call draw_func\n");
	SpPrimitivePart(component).draw_func(component, SpPrimitivePart(component).draw_data);
    }
#else
    spEmitDrawCallback(component);
#endif
    
    spDebug(30, "spDrawImage", "done\n");
    
    return;
}

void spRedrawImage(spComponent component, int width, int height)
{
    if (component == NULL) return;

    spDebug(30, "spRedrawImage", "input size = (%d, %d), client size = (%d, %d)\n",
	    width, height, SpComponentPart(component).client_width, SpComponentPart(component).client_height);

    if ((SpComponentPart(component).client_width != width && width > 0) ||
	(SpComponentPart(component).client_height != height && height > 0)) {
	SpComponentPart(component).client_width = width;
	SpComponentPart(component).client_height = height;

	spRedrawImageArch(component);
    }
    
    spDrawImage(component);
    
    spDebug(80, "spRedrawImage", "done\n");
    
    return;
}

void spResizeImage(spComponent component, int width, int height)
{
    if (component == NULL) return;
    
    if (SpComponentPart(component).client_width != width
	|| SpComponentPart(component).client_height != height) {
	spRedrawImage(component, width, height);
    }
    
    return;
}
