/*
 *	spMenuItem_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <gdk/gdkkeysyms.h>

#include <sp/spBase.h>
#include <sp/spOption.h>

#include <sp/spStatusBar.h>

#include <sp/spFrameP.h>
#include <sp/spButtonP.h>
#include <sp/spMenuItemP.h>

gint spShowMenuHelpCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    spComponent component = (spComponent)data;

    if (component == NULL
	|| (event->type != GDK_ENTER_NOTIFY && event->type != GDK_LEAVE_NOTIFY)) return FALSE;
    
    if (event->type == GDK_ENTER_NOTIFY
	&& SpGetDescription(component) != NULL) {
	spSetHelpStatusText(component, SpGetDescription(component));
    } else {
	spSetHelpStatusText(component, "");
    }

    return FALSE;
}

static void createRadioButtonMenuItem(spComponent component, const char *label)
{
    spComponent prev_radio;
    GSList *group = NULL;

    spAddRadioGroup(component, SpButtonPart(component).radio_end);

    if ((prev_radio = spGetRadioGroupPrev(component)) != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(SpPrimitiveArch(prev_radio).widget));
#else
	group = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(SpPrimitiveArch(prev_radio).widget));
#endif
    }
    
#if GTK_CHECK_VERSION(2,0,0)
    SpPrimitiveArch(component).widget = gtk_radio_menu_item_new_with_mnemonic(group, label);
#else
    SpPrimitiveArch(component).widget = gtk_radio_menu_item_new_with_label(group, label);
#endif
#if GTK_CHECK_VERSION(3,0,0)
    gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(SpPrimitiveArch(component).widget));
#else
    gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(SpPrimitiveArch(component).widget));
#endif

    return;
}

void spMenuItemCreateArch(spComponent component)
{
    static char label[SP_MAX_MENU_LABEL] = "";

    strcpy(label, "");
    if (!strnone(SpComponentPart(component).title)) {
	spGetMnemonicGtk(SpComponentPart(component).title, label);
    }
    
    if (strnone(label)) {
	strcpy(label, (!strnone(SpGetName(component)) ? SpGetName(component) : ""));
    }

    if (spIsCheckBoxMenuItem(component) == SP_TRUE) {
#if GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).widget = gtk_check_menu_item_new_with_mnemonic(label);
#else
	SpPrimitiveArch(component).widget = gtk_check_menu_item_new_with_label(label);
#endif
    } else if (spIsRadioButtonMenuItem(component) == SP_TRUE) {
	createRadioButtonMenuItem(component, label);
    } else {
#if GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).widget = gtk_menu_item_new_with_mnemonic(label);
#else
	SpPrimitiveArch(component).widget = gtk_menu_item_new_with_label(label);
#endif
    }
#if GTK_CHECK_VERSION(3,0,0)
    gtk_menu_shell_append(GTK_MENU_SHELL(SpParentPrimitiveArch(component).widget),
			  SpPrimitiveArch(component).widget);
#else
    gtk_menu_append(GTK_MENU(SpParentPrimitiveArch(component).widget),
		    SpPrimitiveArch(component).widget);
#endif
    gtk_widget_show(SpPrimitiveArch(component).widget);

    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
			    "enter_notify_event", spShowMenuHelpCB,
			    (gpointer)component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
			    "leave_notify_event", spShowMenuHelpCB,
			    (gpointer)component);

    spDebug(40, "spMenuItemCreateArch", "done\n");
    
    return;
}

void spMenuItemSetParamsArch(spComponent component)
{
    static char label[SP_MAX_BUTTON_LABEL] = "";

    if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {    
	strcpy(label, "");
	if (!strnone(SpComponentPart(component).title)) {
	    spGetMnemonicGtk(SpComponentPart(component).title, label);
	}
	
	if (!strnone(label)) {
	    spUpdateWidgetTitleGtk(SpPrimitiveArch(component).widget, label);
	}
    }

    return;
}

void spMenuSeparatorCreateArch(spComponent component)
{
    SpPrimitiveArch(component).widget = gtk_menu_item_new();
    gtk_container_add(GTK_CONTAINER(SpParentPrimitiveArch(component).widget),
		      SpPrimitiveArch(component).widget);
    gtk_widget_show(SpPrimitiveArch(component).widget);
    
    return;
}

static spBool convertKeySym(const char *keyname, guint *key, char *keylabel)
{
    int len;
    spBool flag = SP_TRUE;

    if (key == NULL || strnone(keyname)) {
	flag = SP_FALSE;
    } else {
	len = strlen(keyname);
	if (streq(keyname, "RET") ||
	    streq(keyname, "RETURN") || streq(keyname, "Return") ||
	    streq(keyname, "ENTER") || streq(keyname, "Enter")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Return;
#else
	    *key = GDK_Return;
#endif
	    strcpy(keylabel, "Return");
	} else if (streq(keyname, "ESC") ||
		   streq(keyname, "ESCAPE") || streq(keyname, "Escape")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Escape;
#else
	    *key = GDK_Escape;
#endif
	    strcpy(keylabel, "Escape");
	} else if (streq(keyname, "SPC") ||
		   streq(keyname, "SPACE") || streq(keyname, "Space")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_space;
#else
	    *key = GDK_space;
#endif
	    strcpy(keylabel, "Space");
	} else if (streq(keyname, "BACK") || streq(keyname, "BackSpace") ||
		   streq(keyname, "BACKSPACE") || streq(keyname, "Backspace")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_BackSpace;
#else
	    *key = GDK_BackSpace;
#endif
	    strcpy(keylabel, "BackSpace");
	} else if (streq(keyname, "DEL") ||
		   streq(keyname, "DELETE") || streq(keyname, "Delete")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Delete;
#else
	    *key = GDK_Delete;
#endif
	    strcpy(keylabel, "Delete");
	} else if (streq(keyname, "INS") ||
		   streq(keyname, "INSERT") || streq(keyname, "Insert")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Insert;
#else
	    *key = GDK_Insert;
#endif
	    strcpy(keylabel, "Insert");
	} else if (streq(keyname, "TAB") || streq(keyname, "Tab")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Tab;
#else
	    *key = GDK_Tab;
#endif
	    strcpy(keylabel, "Tab");
	} else if (streq(keyname, "LEFT") || streq(keyname, "Left")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Left;
#else
	    *key = GDK_Left;
#endif
	    strcpy(keylabel, "Left");
	} else if (streq(keyname, "UP") || streq(keyname, "Up")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Up;
#else
	    *key = GDK_Up;
#endif
	    strcpy(keylabel, "Up");
	} else if (streq(keyname, "RIGHT") || streq(keyname, "Right")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Right;
#else
	    *key = GDK_Right;
#endif
	    strcpy(keylabel, "Right");
	} else if (streq(keyname, "DOWN") || streq(keyname, "Down")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Down;
#else
	    *key = GDK_Down;
#endif
	    strcpy(keylabel, "Down");
	} else if (streq(keyname, "HOME") || streq(keyname, "Home")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Home;
#else
	    *key = GDK_Home;
#endif
	    strcpy(keylabel, "Home");
	} else if (streq(keyname, "END") || streq(keyname, "End")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_End;
#else
	    *key = GDK_End;
#endif
	    strcpy(keylabel, "End");
	} else if (streq(keyname, "PRIOR") || streq(keyname, "Prior")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Prior;
#else
	    *key = GDK_Prior;
#endif
	    strcpy(keylabel, "Prior");
	} else if (streq(keyname, "NEXT") || streq(keyname, "Next")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_Next;
#else
	    *key = GDK_Next;
#endif
	    strcpy(keylabel, "Next");
	} else if (streq(keyname, "F1")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F1;
#else
	    *key = GDK_F1;
#endif
	    strcpy(keylabel, "F1");
	} else if (streq(keyname, "F2")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F2;
#else
	    *key = GDK_F2;
#endif
	    strcpy(keylabel, "F2");
	} else if (streq(keyname, "F3")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F3;
#else
	    *key = GDK_F3;
#endif
	    strcpy(keylabel, "F3");
	} else if (streq(keyname, "F4")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F4;
#else
	    *key = GDK_F4;
#endif
	    strcpy(keylabel, "F4");
	} else if (streq(keyname, "F5")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F5;
#else
	    *key = GDK_F5;
#endif
	    strcpy(keylabel, "F5");
	} else if (streq(keyname, "F6")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F6;
#else
	    *key = GDK_F6;
#endif
	    strcpy(keylabel, "F6");
	} else if (streq(keyname, "F7")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F7;
#else
	    *key = GDK_F7;
#endif
	    strcpy(keylabel, "F7");
	} else if (streq(keyname, "F8")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F8;
#else
	    *key = GDK_F8;
#endif
	    strcpy(keylabel, "F8");
	} else if (streq(keyname, "F9")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F9;
#else
	    *key = GDK_F9;
#endif
	    strcpy(keylabel, "F9");
	} else if (streq(keyname, "F10")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F10;
#else
	    *key = GDK_F10;
#endif
	    strcpy(keylabel, "F10");
	} else if (streq(keyname, "F11")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F11;
#else
	    *key = GDK_F11;
#endif
	    strcpy(keylabel, "F11");
	} else if (streq(keyname, "F12")) {
#if GTK_CHECK_VERSION(3,0,0)
	    *key = GDK_KEY_F12;
#else
	    *key = GDK_F12;
#endif
	    strcpy(keylabel, "F12");
	} else if (len == 1) {
	    *key = toupper(keyname[0]);
	    /*sprintf(keylabel, "%c", toupper(keyname[0]));*/
	    sprintf(keylabel, "%c", *key);
	} else {
	    spWarning("unknown key symbol: %s\n", keyname);
	    flag = SP_FALSE;
	}
    }

    spDebug(80, "convertKeySym", "keyname = %s, key = %c, keylabel = %s\n",
	    keyname, *key, keylabel);
    
    return flag;
}

static spBool getAccelKey(const char *shortcut, guint *key, GdkModifierType *mods)
{
    spBool flag = SP_TRUE;
    char *p = NULL;
    static char keylabel[SP_MAX_LINE] = "";

    if (!strnone(shortcut)) {
	*mods = 0;
	if ((p = strchr(shortcut, '-')) != NULL) {
	    switch (shortcut[0]) {
	      case 'C':
		*mods = GDK_CONTROL_MASK;
		break;
	      case 'S':
		*mods = GDK_SHIFT_MASK;
		break;
	      case 'M':
	      case 'A':
		*mods = GDK_MOD1_MASK;
		break;
	      default:
		flag = SP_FALSE;
		break;
	    }
	    if (flag == SP_TRUE &&
		convertKeySym(p + 1, key, keylabel) == SP_TRUE) {
		flag = SP_TRUE;
	    } else {
		spMessage("unknown shortcut: %s\n", shortcut);
		flag = SP_FALSE;
	    }
	} else if (convertKeySym(shortcut, key, keylabel) == SP_TRUE) {
	    flag = SP_TRUE;
	} else {
	    flag = SP_FALSE;
	}
    } else {
	flag = SP_FALSE;
    }
    
    return flag;
}

spBool spAddShortcutArch(spComponent component, const char *shortcut)
{
    spComponent window;
    guint accel_key;
    GdkModifierType accel_mods;

    if (SpPrimitiveArch(component).widget == NULL) return SP_FALSE;
    
    spDebug(80, "spAddShortcutArch", "shortcut = %s\n", shortcut);
    
    window = SpGetWindow(component);

    if (getAccelKey(shortcut, &accel_key, &accel_mods) == SP_TRUE) {
	spDebug(80, "spAddShortcutArch", "accel_key = %d\n", accel_key);
#if GTK_CHECK_VERSION(1,2,0)
	gtk_widget_add_accelerator(SpPrimitiveArch(component).widget,
				   "activate",
				   SpFrameArch(window).accel_group,
				   accel_key,
				   accel_mods,
				   GTK_ACCEL_VISIBLE);
#else
	gtk_widget_install_accelerator(SpPrimitiveArch(component).widget,
				       SpFrameArch(window).accel_group,
				       "activate",
				       accel_key,
				       accel_mods);
#endif
	return SP_TRUE;
    }
    
    spDebug(80, "spAddShortcutArch", "getAccelKey failed\n");
    
    return SP_FALSE;
}
