/*
 *	spMenu_Mac.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spFrameP.h>
#include <sp/spMenuItemP.h>
#include <sp/spMenuP.h>

extern spTopLevel sp_toplevel;

void spSetMenuItemTextMac(MenuHandle menu, int index, Str255 pstr)
{
    int i;
    
    if (pstr[0] <= 0) {
	pstr[0] = 1; pstr[1] = NUL;
    } else if (pstr[1] == '-') {
	for (i = pstr[0] + 1; i >= 2; i--) {
	    pstr[i] = pstr[i - 1];
	}
	pstr[1] = NUL;
	pstr[0] += 1;
    }

    SetMenuItemText(menu, index, pstr);

    return;
}

void spGetMenuItemTextMac(MenuHandle menu, int index, Str255 pstr)
{
    int i;
    
    GetMenuItemText(menu, index, pstr);
    if (pstr[0] > 1 && pstr[1] == NUL) {
	for (i = 1; i < pstr[0]; i++) {
	    pstr[i] = pstr[i + 1];
	}
	pstr[0] -= 1;
    }
    
    return;
}

spBool spGetMenuLabelMac(spComponent component, Str255 plabel)
{
    spBool flag;
    char label[SP_MAX_MENU_LABEL];

    flag = spGetNonMnemonicTitle(component, label);
    spStrCToP(label, plabel);
    
    return flag;
}

void spAddSubMenuMac(spComponent component, Str255 plabel)
{
    int menu_id;
    Str255 pstr;

#if TARGET_API_MAC_CARBON
    spStrCToP(" ", pstr);
#else
    spDebug(80, "spAddSubMenuMac", "%s: component menu_id = %d\n",
	    spGetName(component), SpPrimitiveArch(component).menu_id);
    if (spGetAppearanceVersionMac() >= 0x00000101) {
	spStrCToP(" ", pstr);
    } else {
	char buf[SP_MAX_LINE];

	/* the following method doesn't work on 10.3 or above
	   (the mark indicating submenu is not displayed) */
	sprintf(buf, " !%c/%c", (char)SpPrimitiveArch(component).menu_id,
		'\033');	/* ESC */
	spStrCToP(buf, pstr);
    }
#endif
    
    menu_id = spGetMenuItemIdMac(component);
    spDebug(80, "spAddSubMenuMac", "menu_id = %d\n", menu_id);
    InsertMenuItem(SpParentPrimitiveArch(component).menu, pstr, menu_id);

#if TARGET_API_MAC_CARBON
    SetMenuItemHierarchicalMenu(SpParentPrimitiveArch(component).menu,
				menu_id,
				SpPrimitiveArch(component).menu);
#else
    if (spGetAppearanceVersionMac() >= 0x00000101) {
	SetMenuItemHierarchicalID(SpParentPrimitiveArch(component).menu,
				  menu_id, SpPrimitiveArch(component).menu_id);
    }
#endif
    
    /* prevent from disable of menu including '(' */
    spSetMenuItemTextMac(SpParentPrimitiveArch(component).menu, menu_id, plabel);

    return;
}

void spPulldownMenuCreateArch(spComponent component)
{
    Str255 pstr;

    spGetMenuLabelMac(component, pstr);
    
    if (spIsSubClass(SpGetParent(component), SpMenuBar) == SP_FALSE) {
	SpPrimitiveArch(component).menu_id = spUseCurrentMenuIdMac();
    
	SpPrimitiveArch(component).menu =
	    NewMenu(SpPrimitiveArch(component).menu_id, pstr);

	if (spIsSubClass(component, SpPopupMenu) == SP_TRUE) {
	    SpParentPrimitiveArch(component).popup_menu = component;
	} else {
	    if (SpParentMenuPart(component).help_flag == SP_FALSE) { /* exclude help menu */
		spAddSubMenuMac(component, pstr);
	    }
	}
	/* InsertMenu is called in window selecting to avoid shortcut conflict */
	if (IsWindowVisible(SpPrimitiveArch(SpGetWindow(component)).window)) {
	    InsertMenu(SpPrimitiveArch(component).menu, -1);
	}
    } else {
	if (SpMenuPart(component).help_flag == SP_TRUE) {
	    SpFrameArch(SpGetWindow(component)).help_menu = component;
	    SpPrimitiveArch(component).menu_id = kHMHelpMenuID;
	    SpPrimitiveArch(component).menu = SpTopLevelArch(sp_toplevel).help_menu;
	} else {
	    SpPrimitiveArch(component).menu_id = spUseCurrentMenuIdMac();
	    SpPrimitiveArch(component).menu =
		NewMenu(SpPrimitiveArch(component).menu_id, pstr);
	}
    }
    
    return;
}

void spPulldownMenuSetParamsArch(spComponent component)
{
    Str255 pstr;

    if (spIsSubClass(SpGetParent(component), SpMenuBar) == SP_FALSE) {
	if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {    
	    if (spGetMenuLabelMac(component, pstr) == SP_TRUE) {
		spSetMenuItemTextMac(SpParentPrimitiveArch(component).menu,
				     SpPrimitiveArch(component).menu_id, pstr);
		DrawMenuBar();
	    }
	}
    }
    
    return;
}

void spPopupMenuCreateArch(spComponent component)
{
    spPulldownMenuCreateArch(component);
    return;
}

void spPopupMenuSetParamsArch(spComponent component)
{
    spPulldownMenuSetParamsArch(component);
    return;
}

void spMenuBarCreateArch(spComponent component)
{
    return;
}
