/*
 *        spSlider_Win.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spTabBox.h>
#include <sp/spSliderP.h>

#include <uxtheme.h>
#include <vssym32.h>

#if !defined(TBS_DOWNISLEFT)
#define TBS_DOWNISLEFT 0x0400
#define TBS_REVERSED 0x0200
#endif

extern spTopLevel sp_toplevel;

int spGetScrollCoefWin(long maximum)
{
    if (maximum >= 32768) {
        return (int)ceil((double)maximum / 32768.0);
    } else {
        return 1;
    }
}

static int getScrollCoef(spComponent component)
{
    return spGetScrollCoefWin(SpSliderPart(component).maximum);
}

void spSliderCreateArch(spComponent component)
{
    SCROLLINFO sinfo;

    SpSliderPart(component).scroll_coef = getScrollCoef(component);

    sinfo.cbSize = sizeof(SCROLLINFO);
    sinfo.fMask = SIF_POS | SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
    sinfo.nMin = SpSliderPart(component).minimum / SpSliderPart(component).scroll_coef;
    sinfo.nMax = SpSliderPart(component).maximum / SpSliderPart(component).scroll_coef;
    sinfo.nPage = SpSliderPart(component).page_size / SpSliderPart(component).scroll_coef + 1;
    sinfo.nPos = SpSliderPart(component).value / SpSliderPart(component).scroll_coef;
    sinfo.nTrackPos = 0;
    spDebug(80, "spSliderCreateArch", "nMin = %d, nMax = %d, nPos = %d\n",
            sinfo.nMin, sinfo.nMax, sinfo.nPos);

    /* create scroll bar */
    SpPrimitiveArch(component).hwnd =
        CreateWindow("SCROLLBAR",
                     (!strnone(SpGetName(component)) ? SpGetName(component) : ""),
                     (SpComponentPart(component).orientation == SP_HORIZONTAL ? SBS_HORZ : SBS_VERT)
                     | WS_CHILD | spGetVisibleStyleWin(component),
                     SpComponentPart(component).x, SpComponentPart(component).y,
                     SpComponentPart(component).current_width,
                     SpComponentPart(component).current_height,
                     SpParentPrimitiveArch(component).hwnd,
                     (HMENU)SpComponentPart(component).component_id,
                     SpTopLevelArch(sp_toplevel).hThisInst,
                     NULL);
    SendMessage(SpPrimitiveArch(component).hwnd, WM_SETFONT,
                (WPARAM)SpTopLevelArch(sp_toplevel).sys_font, MAKELPARAM(TRUE, 0));

    SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_CTL, &sinfo, TRUE);
    
    return;
}

void spSliderSetParamsArch(spComponent component)
{
    SCROLLINFO sinfo;

    if (spIsSensitive(component) == SP_FALSE) return;
    
    spDebug(80, "spSliderSetParamsArch", "in\n");

    SpSliderPart(component).scroll_coef = getScrollCoef(component);
    
    sinfo.cbSize = sizeof(SCROLLINFO);
    sinfo.fMask = SIF_POS | SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
    sinfo.nMin = SpSliderPart(component).minimum / SpSliderPart(component).scroll_coef;
    sinfo.nMax = SpSliderPart(component).maximum / SpSliderPart(component).scroll_coef;
    sinfo.nPage = SpSliderPart(component).page_size / SpSliderPart(component).scroll_coef + 1;
    sinfo.nPos = SpSliderPart(component).value / SpSliderPart(component).scroll_coef;
    sinfo.nTrackPos = 0;
    spDebug(80, "spSliderSetParamsArch", "nMin = %d, nMax = %d, nPos = %d\n",
            sinfo.nMin, sinfo.nMax, sinfo.nPos);

    SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_CTL, &sinfo, TRUE);
    
    spDebug(80, "spSliderSetParamsArch", "done\n");
    
    return;
}

spBool spUpdateHScrollValueWin(int *value, WPARAM wParam, int minimum, int maximum, 
                               int page_size, int increment, int page_increment, int scroll_coef,
                               double zoom_factor, spBool track_call_on)
{
    int new_value;
    UINT sub_message = 0;
    
    sub_message = LOWORD(wParam);
    spDebug(100, "spUpdateHScrollValueWin", "WM_HSCROLL: sub_message = %d, original value = %d\n",
            sub_message, *value);
    switch (sub_message) {
      case SB_LINELEFT:
        if (zoom_factor != 1.0) increment = (int)spRound(zoom_factor * (double)increment);
        *value -= increment;
        break;
      case SB_LINERIGHT:
        if (zoom_factor != 1.0) increment = (int)spRound(zoom_factor * (double)increment);
        *value += increment;
        break;
      case SB_PAGELEFT:
        if (zoom_factor != 1.0) page_increment = (int)spRound(zoom_factor * (double)page_increment);
        *value -= page_increment;
        break;
      case SB_PAGERIGHT:
        if (zoom_factor != 1.0) page_increment = (int)spRound(zoom_factor * (double)page_increment);
        *value += page_increment;
        break;
      case SB_THUMBPOSITION:
      case SB_THUMBTRACK:
        if (sub_message == SB_THUMBTRACK && track_call_on == SP_FALSE) {
            return SP_TRUE;
        }

        /* get current value */
#if 0
        *value = (int)HIWORD(wParam) * scroll_coef;
#else
        new_value = (int)HIWORD(wParam) * scroll_coef;
        spDebug(80, "spUpdateHScrollValueWin", "new_value = %d, *value = %d\n", new_value, *value);
        if (*value == new_value) {
            return SP_TRUE;
        }
        *value = new_value;
#endif
        break;
      default:
        spDebug(80, "spUpdateHScrollValueWin", "return SP_TRUE: sub_message = %d\n", sub_message);
        return SP_TRUE;
    }
    spDebug(100, "spUpdateHScrollValueWin", "WM_HSCROLL: updated value = %d\n", *value);

    return SP_FALSE;
}

spBool spUpdateVScrollValueWin(int *value, WPARAM wParam, int minimum, int maximum, 
                               int page_size, int increment, int page_increment, int scroll_coef,
                               double zoom_factor, spBool track_call_on)
{
    int new_value;
    UINT sub_message = 0;
    
    sub_message = LOWORD(wParam);
    spDebug(100, "spUpdateVScrollValueWin", "WM_VSCROLL: sub_message = %d, original value = %d\n", 
            sub_message, *value);
    switch (sub_message) {
      case SB_LINEUP:
        if (zoom_factor != 1.0) increment = (int)spRound(zoom_factor * (double)increment);
        *value -= increment;
        break;
      case SB_LINEDOWN:
        if (zoom_factor != 1.0) increment = (int)spRound(zoom_factor * (double)increment);
        *value += increment;
        break;
      case SB_PAGEUP:
        if (zoom_factor != 1.0) page_increment = (int)spRound(zoom_factor * (double)page_increment);
        *value -= page_increment;
        break;
      case SB_PAGEDOWN:
        if (zoom_factor != 1.0) page_increment = (int)spRound(zoom_factor * (double)page_increment);
        *value += page_increment;
        break;
      case SB_THUMBPOSITION:
      case SB_THUMBTRACK:
        if (sub_message == SB_THUMBTRACK && track_call_on == SP_FALSE) {
            return SP_TRUE;
        }

        /* get current value */
#if 0
        *value = (int)HIWORD(wParam) * scroll_coef;
#else
        new_value = (int)HIWORD(wParam) * scroll_coef;
        if (*value == new_value) {
            return SP_TRUE;
        }
        *value = new_value;
#endif
        break;
      default:
        spDebug(80, "spUpdateVScrollValueWin", "return SP_TRUE: sub_message = %d\n", sub_message);
        return SP_TRUE;
    }
    spDebug(100, "spUpdateVScrollValueWin", "WM_VSCROLL: updated value = %d\n", *value);

    return SP_FALSE;
}

spBool spUpdateScrollValueWin(int *value, UINT message, WPARAM wParam, int minimum, int maximum, 
                              int page_size, int increment, int page_increment, int scroll_coef,
                              double zoom_factor, spBool track_call_on)
{
    if (message == WM_HSCROLL) {
        if (spUpdateHScrollValueWin(value, wParam, minimum, maximum, page_size, increment,
                                    page_increment, scroll_coef, zoom_factor, track_call_on) == SP_TRUE) {
            return SP_TRUE;
        }
    } else {
        if (spUpdateVScrollValueWin(value, wParam, minimum, maximum, page_size, increment,
                                    page_increment, scroll_coef, zoom_factor, track_call_on) == SP_TRUE) {
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

void spUpdateScrollValueInfoWin(SCROLLINFO *sinfo, int *value,
                                int minimum, int maximum, int page_size, int scroll_coef)
{
    sinfo->cbSize = sizeof(SCROLLINFO);
    sinfo->fMask = SIF_POS | SIF_DISABLENOSCROLL;
    
    *value = MAX(*value, minimum);
    *value = MIN(*value, maximum - page_size);
    sinfo->nPos = *value / scroll_coef;

    return;
}

void spSetSliderValueWin(spComponent component)
{
    SCROLLINFO sinfo;

    if (spIsSensitive(component) == SP_FALSE) return;

    spUpdateScrollValueInfoWin(&sinfo, &SpSliderPart(component).value,
                               SpSliderPart(component).minimum, SpSliderPart(component).maximum,
                               SpSliderPart(component).page_size, SpSliderPart(component).scroll_coef);
        
    if (spIsSlider(component) == SP_TRUE) {
        spDebug(50, "spSetSliderValueWin","min = %d, max = %d, value = %d\n",
                SpSliderPart(component).minimum, SpSliderPart(component).maximum,
                SpSliderPart(component).value);
        SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_CTL, &sinfo, 1);
    } else {
        if (SpPrimitiveArch(component).message == WM_HSCROLL) {
            SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_HORZ, &sinfo, 1);
        } else {
            SetScrollInfo(SpPrimitiveArch(component).hwnd, SB_VERT, &sinfo, 1);
        }
    }
    
    return;
}

void spSetSliderValueArch(spComponent component)
{
    if (spIsSubClass(component, SpTrackBar) == SP_TRUE) {
        spSetTrackBarValueWin(component);
    } else {
        spSetSliderValueWin(component);
    }
    
    return;
}

int spGetSliderValueArch(spComponent component)
{
    return SpSliderPart(component).value;
}

void spDrawTrackBarValueWin(spComponent component, HDC hdc)
{
    int width, height;
    int value;
    double factor = 1.0;
    char format[SP_MAX_LINE];
    SIZE size;
    int textx, texty;
    static char string[SP_MAX_LINE];
    spComponent ancestor_tabbox = NULL;
    HANDLE htabtheme = NULL;

    spDebug(100, "spDrawTrackBarValueWin", "name = %s\n", spGetName(component));
        
    if ((ancestor_tabbox = spGetAncestorTabBox(component)) != NULL
        && spIsOSWinXPLater() == SP_TRUE) {
        if ((htabtheme = OpenThemeData(SpPrimitiveArch(ancestor_tabbox).hwnd, L"TAB")) == NULL) {
            spDebug(80, "spDrawTrackBarValueWin", "OpenThemeData failed\n");
        }
    }
        
    SelectObject(hdc, SpTopLevelArch(sp_toplevel).sys_font);
        
    value = SpSliderPart(component).value;

    if (SpSliderPart(component).decimal_points > 0) {
        sprintf(format, " %%6.%df", SpSliderPart(component).decimal_points);
        factor = pow(10.0, (double)SpSliderPart(component).decimal_points);
        sprintf(string, format, (double)value / factor);
    } else {
        sprintf(string, " %5d", value);
    }
    spDebug(80, "spDrawTrackBarValueWin", "string = %s\n", string);
    spGetSize(component, &width, &height);
    GetTextExtentPoint32(hdc, string, (int)strlen(string), &size);
    if (SpComponentPart(component).orientation == SP_HORIZONTAL) {
        textx = width / 2 - size.cx / 2 - spScaleX(5);
        texty = height - size.cy - spScaleY(3);
    } else {
        textx = width - size.cx- /*2*/spScaleX(2);
        texty = height / 2 - size.cy / 2;
    }
        
    if (htabtheme != NULL) {
        RECT rect;
        SetBkMode(hdc, TRANSPARENT);
        rect.left = textx;
        rect.right = textx + size.cx;
        rect.top = texty;
        rect.bottom = texty + size.cy;
        DrawThemeBackground(htabtheme, hdc, TABP_BODY, 0, &rect, NULL);
        CloseThemeData(htabtheme);
        htabtheme = NULL;
    } else {
        SetBkColor(hdc, GetSysColor(COLOR_BTNFACE));
        SetBkMode(hdc, OPAQUE);
    }

    TextOut(hdc, textx, texty, string, (int)strlen(string));

    return;
}

void spSetTrackBarValueWin(spComponent component)
{
    SpSliderPart(component).value =
        MAX(SpSliderPart(component).value, SpSliderPart(component).minimum);
    SpSliderPart(component).value =
        MIN(SpSliderPart(component).value, SpSliderPart(component).maximum);
    spDebug(50, "spSetTrackBarValueWin", "min = %d, max = %d, value = %d\n",
            SpSliderPart(component).minimum, SpSliderPart(component).maximum,
            SpSliderPart(component).value);

    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETPOS, (WPARAM)TRUE,
                (LPARAM)(spReverseSliderValue(component, SpSliderPart(component).value)
                         / SpSliderPart(component).scroll_coef));

    if (SpSliderPart(component).show_value == SP_TRUE) {
        HDC hdc;
    
        hdc = GetDC(SpPrimitiveArch(component).hwnd);
        spDrawTrackBarValueWin(component, hdc);
        ReleaseDC(SpPrimitiveArch(component).hwnd, hdc);
    }
    
    return;
}

void spTrackBarCreateArch(spComponent component)
{
    DWORD dwOtherStyle = 0;
    int pos;
    
    SpSliderPart(component).scroll_coef = getScrollCoef(component);
    
    /* create track bar */
    SpPrimitiveArch(component).hwnd =
        CreateWindowEx(0,
                       TRACKBAR_CLASS,
                       (!strnone(SpGetName(component)) ? SpGetName(component) : ""),
                       (SpComponentPart(component).orientation == SP_HORIZONTAL ? TBS_HORZ : TBS_VERT)
                       | (SpSliderPart(component).show_scale == SP_TRUE ? TBS_AUTOTICKS : TBS_NOTICKS)
                       | WS_CHILD | WS_TABSTOP | dwOtherStyle | spGetVisibleStyleWin(component),
                       SpComponentPart(component).x, SpComponentPart(component).y,
                       SpComponentPart(component).current_width,
                       SpComponentPart(component).current_height,
                       SpParentPrimitiveArch(component).hwnd,
                       (HMENU)SpComponentPart(component).component_id,
                       SpTopLevelArch(sp_toplevel).hThisInst,
                       NULL);
    SendMessage(SpPrimitiveArch(component).hwnd, WM_SETFONT,
                (WPARAM)SpTopLevelArch(sp_toplevel).sys_font, MAKELPARAM(TRUE, 0));

    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETTICFREQ,
                (WPARAM)(SpSliderPart(component).page_increment / SpSliderPart(component).scroll_coef),
                (LPARAM)1);
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG((SpSliderPart(component).minimum / SpSliderPart(component).scroll_coef),
                                 (SpSliderPart(component).maximum / SpSliderPart(component).scroll_coef)));

    /* This SendMessage is required. Maybe Windows bug. */
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETPOS, (WPARAM)FALSE,
                (LPARAM)(SpSliderPart(component).minimum / SpSliderPart(component).scroll_coef));
    
    pos = spReverseSliderValue(component, SpSliderPart(component).value) / SpSliderPart(component).scroll_coef;
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)pos);
    spDebug(50, "spTrackBarCreateArch", "pos = %ld\n", pos);
    
    return;
}

void spTrackBarSetParamsArch(spComponent component)
{
    int pos;
    
    SpSliderPart(component).scroll_coef = getScrollCoef(component);
    
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETTICFREQ,
                (WPARAM)(SpSliderPart(component).page_increment / SpSliderPart(component).scroll_coef),
                (LPARAM)1);
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG((SpSliderPart(component).minimum / SpSliderPart(component).scroll_coef),
                                 (SpSliderPart(component).maximum / SpSliderPart(component).scroll_coef)));

    pos = spReverseSliderValue(component, SpSliderPart(component).value) / SpSliderPart(component).scroll_coef;
    SendMessage(SpPrimitiveArch(component).hwnd, TBM_SETPOS, (WPARAM)TRUE, (LPARAM)pos);
    spDebug(50, "spTrackBarSetParamsArch", "pos = %ld\n", pos);
    
    return;
}
