/*
 *	spTopLevel_Mac.c
 */
#if defined(MACOSX)
#include <Carbon/Carbon.h>
#else
#include <Gestalt.h>
#include <MacTypes.h>
#include <Quickdraw.h>
#include <MacWindows.h>
#include <Dialogs.h>
#include <Menus.h>
#include <Devices.h>
#include <Events.h>
#include <ToolUtils.h>
#include <Controls.h>
#include <ControlDefinitions.h>
#include <Devices.h>
#include <LowMem.h>
#endif

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spThread.h>

#include <sp/spGraphicsP.h>
#include <sp/spComponentP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spMenuP.h>
#include <sp/spFrameP.h>
#include <sp/spButtonP.h>
#include <sp/spSliderP.h>
#include <sp/spTopLevelP.h>

#if !TARGET_API_MAC_CARBON
#if !defined(THINK_C) && !defined(__MWERKS__)
QDGlobals qd;
#endif
#endif

static spBool sp_exec_from_other_app = SP_TRUE;

#if defined(MACOSX)
#undef SP_USE_EVENT_LOOP_TIMER

#include <pthread.h>

static pthread_t sp_main_thread = NULL;
static void *sp_main_mutex = NULL;
static void *sp_port_mutex = NULL;
static void *sp_draw_mutex = NULL;
static void *sp_control_mutex = NULL;

/*#define DISPATCH_MUTEX_LOCK*/
#undef DISPATCH_MUTEX_LOCK
#endif

#if !TARGET_API_MAC_CARBON
static int sp_num_installed_timer = 0;
#endif

#if 0
#define SP_USE_DEBUG_WINDOW
#else
#undef SP_USE_DEBUG_WINDOW
#endif

#if 0
#define SP_USE_DEBUG_FILE_OUT
#else
#undef SP_USE_DEBUG_FILE_OUT
#endif
#define SP_DEBUG_FILE_ALWAYS_FOPEN

#ifdef SP_USE_EVENT_LOOP_TIMER
EventLoopTimerRef sp_main_run_loop_timer = NULL;
EventLoopTimerRef sp_yield_timer = NULL;
#endif

#if defined(SP_USE_DEBUG_WINDOW)
static WindowPtr sp_debug_window = NULL;

#define SP_DEBUG_WINDOW_LEFT 100
#define SP_DEBUG_WINDOW_TOP 100
#define SP_DEBUG_WINDOW_WIDTH 700
#define SP_DEBUG_WINDOW_HEIGHT 450

#define SP_DEBUG_WINDOW_VMARGIN 20
#define SP_DEBUG_WINDOW_HMARGIN 20
#define SP_DEBUG_WINDOW_LINE_SKIP 20

void spShowDebugMessage(char *message)
{
    Rect rect;
    Str255 pstr;
    GrafPtr save_port;
    static int current_top = SP_DEBUG_WINDOW_VMARGIN;
    
    if (sp_debug_window != NULL) {
	spLockWindowPort(sp_debug_window, &save_port);

	if (current_top >= SP_DEBUG_WINDOW_HEIGHT - SP_DEBUG_WINDOW_VMARGIN) {
	    spGetPortRectMac(sp_debug_window, &rect);
	    EraseRect(&rect);
	    current_top = SP_DEBUG_WINDOW_VMARGIN;
	}

	MoveTo(SP_DEBUG_WINDOW_HMARGIN, current_top);
	spStrCToP(message, pstr);
	DrawString(pstr);
	current_top += SP_DEBUG_WINDOW_LINE_SKIP;
	
	spUnlockWindowPort(sp_debug_window, save_port);
    }

    return;
}
#elif defined(SP_USE_DEBUG_FILE_OUT)
FILE *sp_debug_fp = NULL;

void spShowDebugMessage(char *message)
{
#ifndef SP_DEBUG_FILE_ALWAYS_FOPEN
    if (sp_debug_fp != NULL) {
	fprintf(sp_debug_fp, message);
	fflush(sp_debug_fp);
    }
#else
    if ((sp_debug_fp = fopen("stdout", "a")) != NULL) {
	fprintf(sp_debug_fp, message);
	fclose(sp_debug_fp);
    }
    sp_debug_fp = NULL;
#endif
    return;
}
#endif

spBool spIsExecFromOtherAppMac(void)
{
    return sp_exec_from_other_app;
}

void spInitializeFromMainMac(void)
{
#if !TARGET_API_MAC_CARBON
    MaxApplZone();
#endif
    sp_exec_from_other_app = SP_FALSE;
    
    return;
}

static void getMenuLabel(const char *label, const char *shortcut, Str255 plabel)
{
    char buf[SP_MAX_MENU_LABEL];
    char string[SP_MAX_MENU_LABEL];

    spGetMnemonic(label, string);
    
    if (shortcut == NULL) {
	strcpy(buf, string);
    } else {
	sprintf(buf, "%s%s", string, shortcut);
    }
    spDebug(50, "getMenulabel", "buf = %s\n", buf);
    
    spStrCToP(buf, plabel);
    return;
}

static void spCreateDefaultMenuMac(spTopLevel toplevel)
{
    int i;
    int num_file_menu;
    Str255 plabel;
    
    SpTopLevelArch(toplevel).menu_bar = GetMenuBar();
    SpTopLevelArch(toplevel).apple_menu = NewMenu(SP_APPLE_MENU_ID, "\p\024");
    getMenuLabel(SP_ABOUT_MENU_LABEL, NULL, plabel);
    AppendMenu(SpTopLevelArch(toplevel).apple_menu, plabel);
    AppendMenu(SpTopLevelArch(toplevel).apple_menu, "\p(-");
#if !TARGET_API_MAC_CARBON
    AppendResMenu(SpTopLevelArch(toplevel).apple_menu, 'DRVR');
#endif
    InsertMenu(SpTopLevelArch(toplevel).apple_menu, 0);

    getMenuLabel(SP_FILE_MENU_LABEL, NULL, plabel);
    SpTopLevelArch(toplevel).file_menu = NewMenu(SP_FILE_MENU_ID, plabel);
    getMenuLabel(SP_NEW_MENU_LABEL, "/N", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    getMenuLabel(SP_OPEN_MENU_LABEL, "/O", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, "\p(-");
    getMenuLabel(SP_CLOSE_MENU_LABEL, "/W", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    getMenuLabel(SP_SAVE_MENU_LABEL, "/S", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    getMenuLabel(SP_SAVE_AS_MENU_LABEL, NULL, plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, "\p(-");
    getMenuLabel(SP_PAGE_SETUP_MENU_LABEL, NULL, plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    getMenuLabel(SP_PRINT_MENU_LABEL, "/P", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);

    SpTopLevelArch(toplevel).window_menu = NULL;
    
#if !TARGET_API_MAC_CARBON
    AppendMenu(SpTopLevelArch(toplevel).file_menu, "\p(-");
    getMenuLabel(SP_QUIT_MENU_LABEL, "/Q", plabel);
    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
    num_file_menu = SP_QUIT_MENU_ITEM_ID;
#else
    {
	if (spIsAquaMac() == SP_TRUE) {
	    num_file_menu = SP_QUIT_MENU_ITEM_ID - 1;
	} else {
	    AppendMenu(SpTopLevelArch(toplevel).file_menu, "\p(-");
	    getMenuLabel(SP_QUIT_MENU_LABEL, "/Q", plabel);
	    AppendMenu(SpTopLevelArch(toplevel).file_menu, plabel);
	    
	    SetMenuItemCommandID(SpTopLevelArch(toplevel).file_menu,
				 SP_QUIT_MENU_ITEM_ID, 'quit');
	    
	    num_file_menu = SP_QUIT_MENU_ITEM_ID;
	}

	if (SpTopLevelPart(toplevel).use_window_menu == SP_TRUE) {
	    if (CreateStandardWindowMenu(0, &SpTopLevelArch(toplevel).window_menu) == noErr) {
		SetMenuID(SpTopLevelArch(toplevel).window_menu, SP_WINDOW_MENU_ID);
		InsertMenu(SpTopLevelArch(toplevel).window_menu, 0);
	    }
	}
    }
#endif
    for (i = 1; i < num_file_menu; i++) {
	DisableMenuItem(SpTopLevelArch(toplevel).file_menu, i);
    }
    
    getMenuLabel(SP_EDIT_MENU_LABEL, NULL, plabel);
    SpTopLevelArch(toplevel).edit_menu = NewMenu(SP_EDIT_MENU_ID, plabel);
    getMenuLabel(SP_UNDO_MENU_LABEL, "/Z", plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, "\p(-");
    getMenuLabel(SP_CUT_MENU_LABEL, "/X", plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    getMenuLabel(SP_COPY_MENU_LABEL, "/C", plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    getMenuLabel(SP_PASTE_MENU_LABEL, "/V", plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    getMenuLabel(SP_DELETE_MENU_LABEL, NULL, plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, "\p(-");
    getMenuLabel(SP_SELECT_ALL_MENU_LABEL, "/A", plabel);
    AppendMenu(SpTopLevelArch(toplevel).edit_menu, plabel);
    DisableMenuItem(SpTopLevelArch(toplevel).edit_menu, SP_UNDO_MENU_ITEM_ID);

#if !TARGET_API_MAC_CARBON
    HMGetHelpMenuHandle(&SpTopLevelArch(toplevel).help_menu);
#else
    HMGetHelpMenu(&SpTopLevelArch(toplevel).help_menu, NULL);
#endif
    SpTopLevelArch(toplevel).num_help_menu = CountMenuItems(SpTopLevelArch(toplevel).help_menu);
    spDebug(50, "spCreateDefaultMenuMac", "num_help_menu = %d\n", SpTopLevelArch(toplevel).num_help_menu);
    
    return;
}

static void spCreateDebugWindowMac(void)
{
    spDebug(50, "spCreateDebugWindowMac", "in\n");
#if defined(SP_USE_DEBUG_WINDOW)
    if (1/*spGetDebugLevel() >= 0*/) {
	Ptr ptr;
	Rect rect;

	spDebug(50, "spCreateDebugWindowMac", "create window\n");
	
#if !TARGET_API_MAC_CARBON
	ptr = NewPtr(sizeof(CWindowRecord));
#else
	ptr = NULL;
#endif
	
	SetRect(&rect,
		SP_DEBUG_WINDOW_LEFT, SP_DEBUG_WINDOW_TOP,
		SP_DEBUG_WINDOW_LEFT + SP_DEBUG_WINDOW_WIDTH,
		SP_DEBUG_WINDOW_TOP + SP_DEBUG_WINDOW_HEIGHT);
	
#if !TARGET_API_MAC_CARBON
	sp_debug_window = NewCWindow(ptr, &rect, "\pDebug", true,
				     zoomDocProc, (WindowPtr)-1, true, 0L);
#else
	sp_debug_window = NULL;
	if (CreateNewWindow(kDocumentWindowClass,
			    kWindowCloseBoxAttribute | kWindowStandardHandlerAttribute,
			    &rect, &sp_debug_window) == noErr) {
	    SetWTitle(sp_debug_window, "\pDebug");
	    ShowWindow(sp_debug_window);
	}
#endif
	if (sp_debug_window != NULL) {
	    spSetPrintFunc(spShowDebugMessage);
	}
	spDebug(50, "spCreateDebugWindowMac", "done\n");
    }
#elif defined(SP_USE_DEBUG_FILE_OUT)
    {
#ifndef SP_DEBUG_FILE_ALWAYS_FOPEN
	if ((sp_debug_fp = fopen("stdout", "w")) != NULL) {
	    spSetPrintFunc(spShowDebugMessage);
	}
#endif
    }
#endif

    return;
}

static GWorldPtr sp_dummy_gworld = NULL;

static void spCreateDummyGWorldMac(void)
{
    Rect rect;
#if TARGET_API_MAC_CARBON
    GWorldFlags flags = 0;
#else
    GWorldFlags flags = useTempMem;
#endif
    
    SetRect(&rect, 0, 0, SP_DUMMY_GWORLD_WIDTH, SP_DUMMY_GWORLD_HEIGHT);
    
    if (noErr != NewGWorld(&sp_dummy_gworld,
			   0, &rect, NULL, NULL, flags)) {
	sp_dummy_gworld = NULL;
    }
    
    return;
}

static void spDestroyDummyGWorldMac(void)
{
    if (sp_dummy_gworld != NULL) {
	DisposeGWorld(sp_dummy_gworld);
    }

    return;
}

spBool spSetDummyGWorldMac(void)
{
    if (sp_dummy_gworld != NULL) {
	SetGWorld(sp_dummy_gworld, NULL);
    }
    return SP_TRUE;
}

spBool spGetOriginalGWorldMac(GWorldPtr *save_gwp, GDHandle *save_gdh)
{
#if 0
    if (sp_dummy_gworld != NULL) {
	*save_gwp = sp_dummy_gworld;
	*save_gdh = GetGWorldDevice(sp_dummy_gworld);
	return SP_TRUE;
    }
#endif
    
    GetGWorld(save_gwp, save_gdh);
    return SP_TRUE;
}

static SInt32 sp_system_version = 0;
static SInt32 sp_appearance_version = 0;
static SInt32 sp_carbon_version = 0;
static spBool sp_is_aqua = SP_FALSE;
static spBool sp_carbon_event_available = SP_FALSE;
static short sp_main_res_file = 0;

void spInitManagersMac(void)
{
    static int flag = 0;
    SInt32 response;

    if (flag == 0) {
	if (spIsExecFromOtherAppMac() == SP_FALSE) {
#if !TARGET_API_MAC_CARBON
	    InitGraf(&qd.thePort);
	    InitFonts();
	    InitWindows();
	    InitMenus();
	    TEInit();
	    InitDialogs(0L);
#endif
	    FlushEvents(everyEvent,0);
	    InitCursor();
	}
	sp_main_res_file = CurResFile();

	if (Gestalt(gestaltSystemVersion, &response) == noErr) {
	    sp_system_version = response;
	}
	if (Gestalt(gestaltAppearanceAttr, &response) == noErr
	    && Gestalt(gestaltAppearanceVersion,&response) == noErr) {
	    sp_appearance_version = response;
	}
#if TARGET_API_MAC_CARBON
	if (Gestalt(gestaltCarbonVersion, &response) == noErr) {
	    sp_carbon_version = response;
	}
	if (Gestalt(gestaltMenuMgrAttr, &response) == noErr) {
	    if (response & gestaltMenuMgrAquaLayoutMask) {
		sp_is_aqua = SP_TRUE;
	    } else {
		sp_is_aqua = SP_FALSE;
	    }
	}
	if (sp_is_aqua == SP_TRUE || sp_carbon_version >= 0x130) {
	    sp_carbon_event_available = SP_TRUE;
	}
#endif

#if defined(MACOSX)
	if (sp_main_thread == NULL) {
	    sp_main_thread = pthread_self();
	}
	if (sp_main_mutex == NULL) {
	    sp_main_mutex = spCreateMutex(NULL);
	}
	if (sp_port_mutex == NULL) {
	    sp_port_mutex = spCreateMutex(NULL);
	}
	if (sp_draw_mutex == NULL) {
	    sp_draw_mutex = spCreateMutex(NULL);
	}
	if (sp_control_mutex == NULL) {
	    sp_control_mutex = spCreateMutex(NULL);
	}

	spSetMacReadableEncoding(CFStringGetSystemEncoding());
#endif

	if (spIsExecFromOtherAppMac() == SP_FALSE) {
	    spCreateDebugWindowMac();
	}
	
	flag = 1;
    }

    return;
}

long spGetSystemVersionMac(void)
{
    return (long)sp_system_version;
}

long spGetAppearanceVersionMac(void)
{
#if 1
    return (long)sp_appearance_version;
#else
    return 0;
#endif
}

long spGetCarbonVersionMac(void)
{
    return (long)sp_carbon_version;
}

spBool spIsCarbonEventAvailableMac(void)
{
#if TARGET_API_MAC_CARBON
    return sp_carbon_event_available;
#else
    return SP_FALSE;
#endif
}

spBool spIsAquaMac(void)
{
    return sp_is_aqua;
}

int spGetPixelDepthMac(Boolean *is_color)
{
    static GDHandle device = NULL;
    int depth;

    if (device == NULL) {
	device = GetGDevice();
    }

    depth = (*(*device)->gdPMap)->pixelSize;
    if (is_color != NULL) {
	*is_color = BitTst(&(*device)->gdFlags,gdDevType);
    }

    return depth;
}

static CursHandle sp_prev_cursor = NULL;

void spSetCursorMac(CursHandle cursor)
{
    if (cursor == NULL) return;
    
    if (cursor != sp_prev_cursor) {
	spLockDrawMutexMac();

#if 0
	spDebug(20, "spSetCursorMac", "cursor = %ld, prev_cursor = %ld\n",
		(long)cursor, (long)sp_prev_cursor);
	if (sp_prev_cursor != NULL) {
	    ReleaseResource((Handle)sp_prev_cursor);
	}
#endif
	
        HLock((Handle)cursor);
	SetCursor(*cursor);
        HUnlock((Handle)cursor);
	sp_prev_cursor = cursor;
	
	spUnlockDrawMutexMac();
    }
    
    return;
}

void spUnsetCursorMac(void)
{
    spLockDrawMutexMac();
#if !TARGET_API_MAC_CARBON
    SetCursor(&qd.arrow);
#else
    {
	Cursor arrow;

	SetCursor(GetQDGlobalsArrow(&arrow));
    }
#endif
    
    if (sp_prev_cursor != NULL) {
#if 0
	ReleaseResource((Handle)sp_prev_cursor);
#endif
	sp_prev_cursor = NULL;
    }
    spUnlockDrawMutexMac();
    
    return;
}

short spSetMainResFileMac(void)
{
    short cur_res;

    cur_res = CurResFile();

    UseResFile(sp_main_res_file);
    
    return cur_res;
}

Handle spGetResourceMac(ResType type, short id)
{
    short cur_res;
    Handle hres;

    cur_res = spSetMainResFileMac();
    hres = Get1Resource(type, id);
    UseResFile(cur_res);

    return hres;
}

Handle spGetNamedResourceMac(ResType type, ConstStr255Param name)
{
    short cur_res;
    Handle hres;

    cur_res = spSetMainResFileMac();
    hres = Get1NamedResource(type, name);
    /*UseResFile(cur_res);*/

    return hres;
}

static spBool sp_done = SP_FALSE;
static spBool sp_started = SP_FALSE;
static int sp_argc = 0;
static char **sp_argv = NULL;
static spMainFunc sp_main_func = NULL;

spBool spIsStartedMac(void)
{
    return sp_started;
}

pascal OSErr spHandleOpenAppEvent(AppleEvent *appEvent, AppleEvent *reply, SInt32 refCon)
{
    if (sp_started == SP_FALSE) {
	sp_started = SP_TRUE;
	if (sp_main_func != NULL) {
	    sp_main_func(sp_argc, sp_argv);
	}
	sp_done = SP_TRUE;
    }
    
    return noErr;
}

pascal OSErr spHandleQuitAppEvent(AppleEvent *appEvent, AppleEvent *reply, SInt32 refCon)
{
    spComponent window;
    
    window = spGetCurrentWindowMac();
    
    if (spIsFrame(window) == SP_TRUE && SpFrameArch(window).quit_menu_item != NULL) {
	spPostMessageMac(SpFrameArch(window).quit_menu_item, SP_ACTIVATE_CALLBACK, SP_CR_ACTIVATE);
    } else {
	spQuit(0);
    }
    
    return noErr;
}

pascal OSErr spHandleOpenDocsEvent(AppleEvent *appEvent, AppleEvent *reply, SInt32 refCon)
{
    int i;
    FSSpec fileSpec;
    AEDescList docList;
    OSErr err;
    SInt32 numItems;
    Size actualSize;
    AEKeyword keyWord;
    DescType returnedType;
    char filename[SP_MAX_PATHNAME];
    int argc;
    char **argv;
    
    if ((err = AEGetParamDesc(appEvent, keyDirectObject, typeAEList, &docList)) == noErr) {
        numItems = 1;
	AECountItems(&docList, &numItems);
	
	if (sp_started == SP_TRUE) {
	    argv = xalloc(numItems, char *);
	    argc = 0;
	} else {
	    if (sp_argv != NULL) {
		argv = xalloc(numItems + sp_argc, char *);
		for (argc = 0; argc < sp_argc; argc++) {
		    argv[argc] = strclone(sp_argv[argc]);
		}
	    } else {
		argv = xalloc(numItems + 1, char *);
		argv[0] = strclone("a.out");
		argc = 1;
	    }
	}
	
	for (i = 1; i <= numItems; i++) {
	    if ((err = AEGetNthPtr(&docList, i, typeFSS, &keyWord, &returnedType,
				   (Ptr)&fileSpec, sizeof(fileSpec), &actualSize)) != noErr) {
		break;
	    }

	    if (spFSSpecToExactNameMac(&fileSpec, filename) == SP_TRUE) {
		argv[argc++] = strclone(filename);
	    }
	}
	
	AEDisposeDesc(&docList);

	if (sp_started == SP_TRUE) {
	    spComponent component;
	    
	    component = spGetCurrentWindowMac();
	    
	    if (spIsFrame(component) == SP_TRUE
		&& SpFrameArch(component).drop_call_func != NULL) {
		/* drop callback */
		SpFrameArch(component).drop_call_func(component, SpFrameArch(component).drop_call_data,
						      argc, argv);
	    }
	} else {
	    sp_started = SP_TRUE;
	    if (sp_main_func != NULL) {
		sp_main_func(argc, argv);
	    }
	    sp_done = SP_TRUE;
	}
	    
	for (i = 0; i < argc; i++) {
	    xfree(argv[i]);
	}
	xfree(argv);
	
	return noErr;
    }
    
    return err;
}

pascal OSErr spHandlePrintDocsEvent(AppleEvent *appEvent, AppleEvent *reply, SInt32 refCon)
{
    return errAEEventNotHandled;
}

static AEEventHandlerUPP spHandleOpenAppEventUPP = NULL;
static AEEventHandlerUPP spHandleQuitAppEventUPP = NULL;
static AEEventHandlerUPP spHandleOpenDocsEventUPP = NULL;
static AEEventHandlerUPP spHandlePrintDocsEventUPP = NULL;
#if TARGET_API_MAC_CARBON
static EventHandlerUPP spAppEventHandlerUPP = NULL;
static EventHandlerRef spAppEventHandlerRef = NULL;

void spInstallApplicationEventHandlerMac(void)
{
    if (spAppEventHandlerRef == NULL && spIsCarbonEventAvailableMac() == SP_TRUE) {
	EventTypeSpec list[] = { 
	    { kEventClassAppleEvent, kEventAppleEvent },
	    { kEventClassMenu, kEventMenuTargetItem },

	    { kEventClassCommand, kEventCommandProcess },
	    { kEventClassMouse, kEventMouseDown },
	    { kEventClassMouse, kEventMouseUp },
	    { kEventClassMouse, kEventMouseMoved },
	    { kEventClassMouse, kEventMouseDragged },
	    { kEventClassMouse, kEventMouseWheelMoved },
#ifdef MACOSX
	    /* spComponent original event */
	    { 'SpCm', 1 },
#endif
	};

	spAppEventHandlerUPP = spNewAppEventHandlerUPP();
	if (spIsExecFromOtherAppMac() == SP_TRUE) {
	    InstallApplicationEventHandler(spAppEventHandlerUPP, GetEventTypeCount(list) - 2, &list[2],
					   0, &spAppEventHandlerRef);
	} else {
	    InstallApplicationEventHandler(spAppEventHandlerUPP, GetEventTypeCount(list), list,
					   0, &spAppEventHandlerRef);
	}
	
	spDebug(20, "spInstallApplicationEventHandlerMac", "installed\n");
    }

    return;
}

void spRemoveApplicationEventHandlerMac(void)
{
    if (spAppEventHandlerRef != NULL) {
	RemoveEventHandler(spAppEventHandlerRef);
	spAppEventHandlerRef = NULL;
    }
    if (spAppEventHandlerUPP != NULL) {
	DisposeEventHandlerUPP(spAppEventHandlerUPP);
	spAppEventHandlerUPP = NULL;
    }

    return;
}
#endif

spBool spInitializeMac(int argc, char **argv, spMainFunc func)
{
    if (argc >= 2 && strneq(argv[1], "-psn_", 5)) {
	argc = 1;
    }
    sp_argc = argc;
    sp_argv = argv;
    
    spDebug(20, "spInitializeMac", "in\n");
    
    spInitManagersMac();

    sp_main_func = func;
    
    spHandleOpenAppEventUPP = NewAEEventHandlerUPP((AEEventHandlerProcPtr)spHandleOpenAppEvent);
    spHandleQuitAppEventUPP = NewAEEventHandlerUPP((AEEventHandlerProcPtr)spHandleQuitAppEvent);
    spHandleOpenDocsEventUPP = NewAEEventHandlerUPP((AEEventHandlerProcPtr)spHandleOpenDocsEvent);
    spHandlePrintDocsEventUPP = NewAEEventHandlerUPP((AEEventHandlerProcPtr)spHandlePrintDocsEvent);
    
    if (AEInstallEventHandler(kCoreEventClass, kAEOpenApplication, spHandleOpenAppEventUPP,
			      0L, false) != noErr) {
	return SP_FALSE;
    }
    if (AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, spHandleQuitAppEventUPP,
			      0L, false) != noErr) {
	return SP_FALSE;
    }
    if (AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments, spHandleOpenDocsEventUPP,
			      0L, false) != noErr) {
	return SP_FALSE;
    }
    if (AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments, spHandlePrintDocsEventUPP,
			      0L, false) != noErr) {
	return SP_FALSE;
    }

#if TARGET_API_MAC_CARBON
    spInstallApplicationEventHandlerMac();
#endif
    
    spDebug(20, "spInitializeMac", "done\n");
    
    return SP_TRUE;
}

void spWaitHighLevelEventMac(void)
{
    EventRecord	event;
    
    sp_done = SP_FALSE;

#if TARGET_API_MAC_CARBON
    if (spIsCarbonEventAvailableMac() == SP_TRUE){
	RunApplicationEventLoop();
	spDebug(20, "spWaitHighLevelEventMac", "RunApplicationEventLoop done\n");
	return;
    }
#endif
    
    while (sp_done == SP_FALSE) {
	if (WaitNextEvent(everyEvent, &event, 0x7FFFFFFF, 0L)) {
	    switch(event.what) {
	      case kHighLevelEvent:
		AEProcessAppleEvent(&event);
		break;
	      default:
		break;
	    }
	}
    }

    return;
}

spBool spIsInPThreadMac(void)
{
#if defined(MACOSX)
    pthread_t current_thread;
    current_thread = pthread_self();
    
    if (current_thread != NULL && !pthread_equal(sp_main_thread, current_thread)) {
	return SP_TRUE;
    }
#endif

    return SP_FALSE;
}

void spLockMainMutexMac(void)
{
#if defined(MACOSX)
    /*spDebug(80, "spLockMainMutexMac", "waiting...\n");*/
    spLockMutex(sp_main_mutex);
    /*spDebug(80, "spLockMainMutexMac", "locked\n");*/
#endif
    return;
}
    
void spUnlockMainMutexMac(void)
{
#if defined(MACOSX)
    spUnlockMutex(sp_main_mutex);
    /*spDebug(80, "spUnlockMainMutexMac", "unlocked\n");*/
#endif
    return;
}

void spLockMainMutexForMainMac(void)
{
#if defined(MACOSX)
    if (spIsInPThreadMac() == SP_FALSE) {
	spLockMainMutexMac();
    }
#endif
    return;
}
    
void spUnlockMainMutexForMainMac(void)
{
#if defined(MACOSX)
    if (spIsInPThreadMac() == SP_FALSE) {
	spUnlockMainMutexMac();
    }
#endif
    return;
}
    
void spLockMainMutexForThreadMac(void)
{
#if defined(MACOSX)
    if (spIsInPThreadMac() == SP_TRUE) {
	/*spDebug(100, "spLockMainMutexForThreadMac", "++++++++ lock for thread ++++++++\n");*/
	spLockMainMutexMac();
    }
#endif
    return;
}
    
void spUnlockMainMutexForThreadMac(void)
{
#if defined(MACOSX)
    if (spIsInPThreadMac() == SP_TRUE) {
	spUnlockMainMutexMac();
	/*spDebug(100, "spUnlockMainMutexForThreadMac", "-------- unlock for thread --------\n");*/
    }
#endif
    return;
}
    
void spLockPortMutexMac(void)
{
#if defined(MACOSX)
    /*spDebug(80, "spLockPortMutexMac", "waiting...\n");*/
    spLockMutex(sp_port_mutex);
    /*spDebug(80, "spLockPortMutexMac", "locked\n");*/
#endif
    return;
}
    
void spUnlockPortMutexMac(void)
{
#if defined(MACOSX)
    spUnlockMutex(sp_port_mutex);
    /*spDebug(80, "spUnlockPortMutexMac", "unlocked\n");*/
#endif
    return;
}
    
void spLockDrawMutexMac(void)
{
#if defined(MACOSX)
    /*spDebug(80, "spLockDrawMutexMac", "waiting...\n");*/
    spLockMutex(sp_draw_mutex);
    /*spDebug(80, "spLockDrawMutexMac", "locked\n");*/
#endif
    return;
}
    
void spUnlockDrawMutexMac(void)
{
#if defined(MACOSX)
    spUnlockMutex(sp_draw_mutex);
    /*spDebug(80, "spUnlockDrawMutexMac", "unlocked\n");*/
#endif
    return;
}
    
void spLockControlMutexMac(void)
{
#if defined(MACOSX)
    /*spDebug(80, "spLockControlMutexMac", "waiting...\n");*/
    spLockMutex(sp_control_mutex);
    /*spDebug(80, "spLockControlMutexMac", "locked\n");*/
#endif
    return;
}
    
void spUnlockControlMutexMac(void)
{
#if defined(MACOSX)
    spUnlockMutex(sp_control_mutex);
    /*spDebug(80, "spUnlockControlMutexMac", "unlocked\n");*/
#endif
    return;
}
    
void spLockMenuMutexMac(void)
{
#if 1
    spLockDrawMutexMac();
    spLockControlMutexMac();
#endif
    return;
}

void spUnlockMenuMutexMac(void)
{
#if 1
    spUnlockControlMutexMac();
    spUnlockDrawMutexMac();
#endif
    return;
}

void spTopLevelPartInitArch(spTopLevel toplevel)
{
    spInitManagersMac();
#if 0
#if TARGET_API_MAC_CARBON
    if (spAppEventHandlerRef == NULL) {
	spInstallApplicationEventHandlerMac();
    }
#endif
#endif
    
    if (spIsExecFromOtherAppMac() == SP_FALSE) {
	SpTopLevelArch(toplevel).message_loop_started = SP_FALSE;
    } else {
	sp_started = SP_TRUE;
	SpTopLevelArch(toplevel).message_loop_started = SP_TRUE;
    }
#ifdef SP_MAC_USER_EVENT_BY_CARBON_EVENT
    SpTopLevelArch(toplevel).event_queue = NULL;
#else
    SpTopLevelArch(toplevel).user_event = NULL;
    spAllocUserEventDefault(&SpTopLevelArch(toplevel).user_event, SP_MAC_DEFAULT_NUM_USER_EVENT);
#endif

#if defined(MACOSX)
    SpTopLevelArch(toplevel).event_loop_event = spCreateEvent(SP_TRUE, SP_TRUE);
#endif
    
#if defined(DISPATCH_MUTEX_LOCK)
    spLockMainMutexMac();
#endif

    return;
}

void spTopLevelPartFreeArch(spTopLevel toplevel)
{
    spDestroyActionProcMac();
    spDestroyDummyGWorldMac();
#if defined(DISPATCH_MUTEX_LOCK)
    spUnlockMainMutexMac();
#endif
    
#if defined(MACOSX)
    if (SpTopLevelArch(toplevel).event_loop_event != NULL) {
	spDestroyEvent(SpTopLevelArch(toplevel).event_loop_event);
    }
#endif
    
#ifndef SP_MAC_USER_EVENT_BY_CARBON_EVENT
    spDestroyAllUserEventDefault(&SpTopLevelArch(toplevel).user_event);
#endif
    
#if defined(MACOSX)
    if (sp_control_mutex != NULL) {
	spDestroyMutex(sp_control_mutex);
	sp_control_mutex = NULL;
    }
    if (sp_draw_mutex != NULL) {
	spDestroyMutex(sp_draw_mutex);
	sp_draw_mutex = NULL;
    }
    if (sp_port_mutex != NULL) {
	spDestroyMutex(sp_port_mutex);
	sp_port_mutex = NULL;
    }
    if (sp_main_mutex != NULL) {
	spDestroyMutex(sp_main_mutex);
	sp_main_mutex = NULL;
    }
#endif
    
#if TARGET_API_MAC_CARBON
    spRemoveApplicationEventHandlerMac();
#endif
    
    return;
}

void spTopLevelCreateArch(spTopLevel toplevel)
{
    spCreateActionProcMac();
    spCreateDummyGWorldMac();
    spCreateDefaultMenuMac(toplevel);
    
#ifdef SP_MAC_USER_EVENT_BY_CARBON_EVENT
    SpTopLevelArch(toplevel).event_queue = GetCurrentEventQueue();
#endif
    
    return;
}

void spTopLevelSetParamsArch(spTopLevel toplevel)
{
    return;
}
    
void spTopLevelDestroyArch(spTopLevel toplevel)
{
    return;
}

void spQuitArch(int status)
{
#if defined(SP_USE_DEBUG_WINDOW)
#elif defined(SP_USE_DEBUG_FILE_OUT)
#ifndef SP_DEBUG_FILE_ALWAYS_FOPEN
    if (sp_debug_fp != NULL) {
	fclose(sp_debug_fp);
    }
#endif
#endif
    
    sp_done = SP_TRUE;
    
    spDebug(20, "spQuitArch", "done\n");
    
    return;
}

void spExitArch(int status)
{
    if (spIsExecFromOtherAppMac() == SP_FALSE) {
	spExit(status);
    }
    return;
}

int spWaitNextEventArch(spTopLevel toplevel, spBool lock_flag)
{
    spDispatchEventMac(toplevel, 0x7FFFFFFF);
    return 1;
}

#if TARGET_API_MAC_CARBON
int spWaitNextEventCarbonMac(unsigned long sleep, EventRef *event)
{
    int flag = 0;
    EventTimeout timeout = kEventDurationNoWait;

    if (sleep <= 0) {
	timeout = kEventDurationNoWait;
    } else if (sleep >= 0x7FFFFFFF) {
	timeout = kEventDurationForever;
    } else {
	timeout = TicksToEventTime(MAX(sleep, 1));
    }

    if (ReceiveNextEvent(0, NULL, timeout, true, event) == noErr) {
	flag = 1;
    }

    return flag;
}

int spDispatchEventCarbonMac(spTopLevel toplevel, unsigned long sleep)
{
    int flag = 0;
    EventTargetRef target;
    EventRef event;

#if defined(DISPATCH_MUTEX_LOCK)
    spUnlockMainMutexMac();
#endif

    flag = spWaitNextEventCarbonMac(sleep, &event);

#if defined(DISPATCH_MUTEX_LOCK)
    spLockMainMutexMac();
#endif

    if (flag) {
	target = GetEventDispatcherTarget();
	SendEventToEventTarget(event, target);
	ReleaseEvent(event);
    }
    
#ifndef SP_USE_EVENT_LOOP_TIMER
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spYieldThread();
    }
#endif
    
    return flag;
}
#endif
    
static UInt32 sp_last_event_modifiers = 0;

UInt32 spGetLastEventModifiersMac(void)
{
#if TARGET_API_MAC_CARBON
    return GetCurrentKeyModifiers();
#else
    return sp_last_event_modifiers;
#endif
}

int spWaitNextEventNonCarbonMac(unsigned long sleep, EventRecord *event)
{
    int flag = 0;
    
    if (WaitNextEvent(everyEvent, event, (UInt32)sleep, 0L)) {
	flag = 1;
    }
    sp_last_event_modifiers = event->modifiers;

    return flag;
}

int spDispatchEventNonCarbonMac(spTopLevel toplevel, unsigned long sleep)
{
    int flag = 0;
    EventRecord	event;

#if defined(DISPATCH_MUTEX_LOCK)
    spUnlockMainMutexMac();
#endif

    flag = spWaitNextEventNonCarbonMac(sleep, &event);
    
#if defined(DISPATCH_MUTEX_LOCK)
    spLockMainMutexMac();
#endif

    if (flag) {
	flag = 0;
	if (spHandleEventMac(&event) == SP_TRUE) {
	    flag = 1;
	}
    } else {
	spHandleIdleMac(&event);
#ifndef SP_MAC_USER_EVENT_BY_CARBON_EVENT
	if (spProcessPendingUserEventDefault(&SpTopLevelArch(toplevel).user_event, 0) > 0) {
	}
#endif
    
    }
    
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spYieldThread();
    }
    
    return flag;
}

int spDispatchEventMac(spTopLevel toplevel, unsigned long sleep)
{
    if (SpTopLevelArch(toplevel).message_loop_started == SP_FALSE) {
	spComponent child;

	child = spGetChild(NULL);

	while (child != NULL) {
	    if (spIsFrame(child) == SP_TRUE
		&& SpFrameArch(child).realize_flag == SP_TRUE) {
		spActivateComponentMac(child, SP_TRUE);
	    }
	    child = SpGetNextComponent(child);
	}
	
	SpTopLevelArch(toplevel).message_loop_started = SP_TRUE;
    }
    
#if TARGET_API_MAC_CARBON
    if (spIsCarbonEventAvailableMac() == SP_TRUE) {
	return spDispatchEventCarbonMac(toplevel, sleep);
    }
#endif

    return spDispatchEventNonCarbonMac(toplevel, sleep);
}
    
int spDispatchEventEffectiveMac(spTopLevel toplevel)
{
    int flag;
    int num_coop_thread;
    
    if (spIsAquaMac() == SP_TRUE) {
	if (spIsCarbonEventAvailableMac() == SP_TRUE) {
	    if (spIsLowLevelThreadUsed() == SP_TRUE) {
		num_coop_thread = 0;
	    } else {
		num_coop_thread = spGetNumCoopThreadMac();
		spDebug(80, "spDispatchEventEffectiveMac", "num_coop_thread = %d\n", num_coop_thread);
	    }
	    
	    if (num_coop_thread <= 0) {
#ifdef SP_USE_EVENT_LOOP_TIMER
		if (sp_yield_timer != NULL) {
		    SetEventLoopTimerNextFireTime(sp_yield_timer, kEventDurationForever);
		}
#endif
		/* no other thread or using pthread */
		flag = spDispatchEventMac(toplevel, 0x7FFFFFFF);
	    } else {
#ifdef SP_USE_EVENT_LOOP_TIMER
		if (sp_yield_timer != NULL) {
		    SetEventLoopTimerNextFireTime(sp_yield_timer, 0.001);
		}
#endif
		flag = spDispatchEventMac(toplevel, 0);
	    }
	} else {
	    flag = spDispatchEventMac(toplevel, 1);
	}
    } else {
	flag = spDispatchEventMac(toplevel, 1);
    }

    return flag;
}

int spDispatchEventArch(spTopLevel toplevel, spBool lock_flag)
{
    int flag;
    
    if (spIsCarbonEventAvailableMac() == SP_TRUE) {
	flag = spDispatchEventMac(toplevel, 0);
    } else {
	flag = spDispatchEventMac(toplevel, 1);
    }

    return flag;
}

static int mainLoop(spTopLevel toplevel)
{
    while (1) {
	spDispatchEventEffectiveMac(toplevel);
    }
    
    return 0;
}

#ifdef SP_USE_EVENT_LOOP_TIMER
static pascal void spYieldTimerFunc(EventLoopTimerRef inTimer, void *inUserData)
{
    spDebug(80, "spYieldTimerFunc", "yield...\n");
    spYieldThread();
    spDebug(80, "spYieldTimerFunc", "done\n");
    
    return;
}

static pascal void spMainRunLoopForThread(EventLoopTimerRef inTimer, void *inUserData)
{
    spTopLevel toplevel = (spTopLevel)inUserData;
    
    spDebug(80, "spMainRunLoopForThread", "in\n");

    mainLoop(toplevel);
    
    spDebug(80, "spMainRunLoopForThread", "done\n");
    
    return;
}
#endif

int spMainLoopArch(spTopLevel toplevel)
{
#ifdef SP_USE_EVENT_LOOP_TIMER
    InstallEventLoopTimer(GetCurrentEventLoop(), 0, 0,
			  NewEventLoopTimerUPP(spYieldTimerFunc),
			  toplevel, &sp_yield_timer);
    if (InstallEventLoopTimer(GetCurrentEventLoop(), 0, 0,
			      NewEventLoopTimerUPP(spMainRunLoopForThread),
			      toplevel, &sp_main_run_loop_timer) == noErr) {
	RunApplicationEventLoop();
	return 0;
    }
#endif
    
    return mainLoop(toplevel);
}

GrafPtr spThreadEnterMac(GrafPtr new_port)
{
    GrafPtr old_port = nil;

#if defined(DISPATCH_MUTEX_LOCK)
    spLockMainMutexMac();
#elif defined(MACOSX)
    spLockMainMutexForThreadMac();
#endif

    if (new_port != nil) {
	GetPort(&old_port);
    
	SetPort(new_port);
    }
    
    return old_port;
}

GrafPtr spThreadLeaveMac(GrafPtr new_port)
{
    GrafPtr old_port = nil;

    if (new_port != nil) {
	GetPort(&old_port);
    
	SetPort(new_port);
    }

#if defined(DISPATCH_MUTEX_LOCK)
    spUnlockMainMutexMac();
#elif defined(MACOSX)
    spUnlockMainMutexForThreadMac();
#endif
    spYieldThread();
    
    return old_port;
}

static GrafPtr sp_old_port = nil;

spBool spThreadEnterArch(spTopLevel toplevel)
{
    sp_old_port = spThreadEnterMac(nil);
    
    return SP_TRUE;
}

spBool spThreadLeaveArch(spTopLevel toplevel)
{
    spThreadLeaveMac(sp_old_port);
    
    return SP_TRUE;
}
    
spBool spSuspendMainLockArch(spTopLevel toplevel)
{
    return SP_TRUE;
}

spBool spWaitThreadPostDummyEventArch(spTopLevel toplevel)
{
    PostEvent(mouseDown, 0);
    PostEvent(mouseUp, 0);
    return SP_TRUE;
}

long spWaitThreadLoopArch(spTopLevel toplevel, void *handle)
{
    return spWaitThreadLoopDefault(toplevel, handle, SP_FALSE);
}

spBool spResumeMainLockArch(spTopLevel toplevel)
{
    return SP_TRUE;
}

spBool spIsPostEventSupportedArch(spTopLevel toplevel)
{
    return SP_TRUE;
}

#if !TARGET_API_MAC_CARBON
spTimerId spSetTimerArch(spTimerList timer)
{
    ThreadBeginCritical();
    timer->last_tick = TickCount();
    ++sp_num_installed_timer;
    ThreadEndCritical();

    return (spTimerId)timer;
}

spBool spKillTimerArch(spTimerList timer)
{
    ThreadBeginCritical();
    timer->last_tick = 0;
    --sp_num_installed_timer;
    ThreadEndCritical();

    return SP_TRUE;
}

int spProcessTimerEventMac(void)
{
    int num_timer = 0;
    spComponent next;
    spBool kill_timer;
    spTimerList timer, prev_timer;
    unsigned long current_tick;
    unsigned long last_tick;
    unsigned long interval_tick;

    if (sp_num_installed_timer > 0) {
	current_tick = TickCount();
	
	next = spGetChild(NULL);
	while (next != NULL) {
	    if (spIsFrame(next) == SP_TRUE) {
		timer = SpFramePart(next).timer;
		while (timer != NULL) {
		    kill_timer = SP_FALSE;
		    
		    last_tick = timer->last_tick;
		    if (timer->call_func != NULL && last_tick > 0) {
			interval_tick = (unsigned long)(60L * timer->interval_ms / 1000L);
			
			if (current_tick - last_tick >= interval_tick) {
			    ++num_timer;
			    
			    if (timer->call_func(timer->frame, timer->call_data) == SP_FALSE) {
				ThreadBeginCritical();
				timer->last_tick = 0;
				kill_timer = SP_TRUE;
				ThreadEndCritical();
			    } else {
				ThreadBeginCritical();
				if (timer->last_tick > 0) {
				    timer->last_tick = current_tick;
				}
				ThreadEndCritical();
			    }
			}
		    }

		    prev_timer = timer;
		    timer = timer->next;

		    if (kill_timer == SP_TRUE) {
			/* 2009/10/1 added */
			spRemoveTimerList(prev_timer->frame, timer);
		    }
		}
	    }
	
	    next = SpGetNextComponent(next);
	}
    }

    return num_timer;
}
#endif
